# This file includes code from:
# LINEX webtool
# Copyright (C) 2021  LipiTUM group, Chair of experimental Bioinformatics, Technical University of Munich
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import warnings
import re
from typing import List, Dict, Tuple, Union, Callable, Any, Set
from itertools import permutations, combinations
from warnings import warn
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib.colorbar import Colorbar
from matplotlib.cm import ScalarMappable
from matplotlib.colors import Normalize, to_hex, to_rgba
import networkx as nx
import networkx.drawing.layout as nx_layout
from pyvis.network import Network
import json

# From package
from .GenerateLipidNetwork import GenerateLipidNetwork
from .parser import (
    parse_fa_reactions,
    parse_fatty_acids,
    parse_excluded_fa_reactions,
    parse_lipid_list
)
from .reference import parse_lipid_reference_table, parse_lipid_reference_table_dict, ReferenceLipid
from .tmp_dirty_functions import (
    make_organism_reaction_list_from_reactome,
    make_all_reaction_list_from_reactome,
    make_all_reaction_list_from_rhea
)
from .lipid import Lipid
from .colour_helper import _generate_colormap_
from .utils import (
    unique_elements,
    SymmetricMatrix,
    extend_class_fa,
    check_databases,
    combine_reactions
)
from .misc import _range_scale_, _size_legend_, _tuple_to_string_, _get_sizes_report_nas_
from .exceptions import NotComputedError, InputDataError
from .vis_utils import VisParser, DynamicVisParser, STATIC_EDGE_PROPERTIES, STATIC_NODE_PROPERTIES
from . import templates
from .load_data import *

with pkg_resources.path(templates, "lipid_colours.json") as path:
    LIPID_CLASS_COLOURS = json.load(open(path, "r"))


class LipidNetwork:
    __slots__ = [
        "gln",
        "layout",
        "network",
        "lipid_mapping",
        "reverse_mapping",
        "last_network_call",
        "_attributes_to_plot",
        "lipid_attributes",
        "interaction_attributes",
        "node_colours",
        "edge_colours",
        "data",
        "groups",
        "unique_groups",
        "reference_group",
        "comparisons",
        "directed",
        "_ref_dict",
        "_incompatible_lipids"
    ]

    network: nx.Graph
    layout: Dict[str, np.ndarray]

    data: pd.DataFrame
    groups: pd.Series
    unique_groups: pd.Series
    comparisons: List[Tuple[str, str]]

    lipid_attributes: Union[Dict[str, Dict[str, Any]], Dict[str, Dict[Union[str, Tuple[str, str]], Dict[str, Any]]]]
    interaction_attributes: Union[
        Dict[str, SymmetricMatrix], Dict[str, Dict[Union[str, Tuple[str, str]], SymmetricMatrix]]]
    node_colours: Dict[Union[str, Tuple[str, str]], Tuple[Union[dict, Dict[str, Union[str, Any]]], list]]
    edge_colours: Dict[Union[str, Tuple[str, str]], Tuple[Union[dict, Dict[str, Union[str, Any]]], list]]
    _attributes_to_plot: Dict[str, Set[str]]

    gln: GenerateLipidNetwork
    _incompatible_lipids: List[str]

    lipid_mapping: Dict[str, List[str]]
    reverse_mapping: Dict[str, str]
    last_network_call: Dict[str, any]

    directed: bool

    # database related slots
    _ref_dict: Dict[str, ReferenceLipid]

    _allowed_node_attributes_ = {
        "lipid_class", "c_index", "db_index", "oh_index",
        "chain_length", "desaturation", "hydroxylation",
        "nlog_pvalues", "fold_changes", "degree",
        "betweenness", "closeness"
    }

    _allowed_edge_attributes_ = {
        "correlations",
        "partial_correlations",
        "reaction_type",
        "enzyme_id",
        "correlation_changes",
        "partial_correlation_changes",
        "enzyme_gene_name",
        "enzyme_uniprot"
    }
    # auxiliary attributes
    _attr_group_size_ = {
        # edge attributes
        "correlations": 1,
        "partial_correlations": 1,
        "reaction_type": 0,
        "enzyme_id": 0,
        "correlation_changes": 2,
        "partial_correlation_changes": 2,
        # node attributes
        "fold_changes": 2,
        "pvalues": 2,
        "nlog_pvalues": 2,
        "lipid_class": 0,
        "c_index": 0,
        "db_index": 0,
        "oh_index": 0,
        "desaturation": 0,
        "hydroxylation": 0,
        "chain_length": 0,
        "closeness": 0,
        "betweenness": 0,
        "degree": 0
    }
    _discrete_map_ = {
        # node features
        "lipid_class": True,
        "c_index": False,
        "db_index": False,
        "oh_index": False,
        "desaturation": False,
        "hydroxylation": False,
        "degree": False,
        "betweenness": False,
        "closeness": False,
        "chain_length": False,
        "fold_changes": False,
        "pvalues": False,
        "nlog_pvalues": False,
        # edge features
        "correlations": False,
        "partial_correlations": False,
        "correlation_changes": True,
        "partial_correlation_changes": True,
        "reaction_type": True
    }

    # Imported methods -------------------------------------------------------------------------------------------------
    # noinspection PyUnresolvedReferences
    from ._network_analysis import (
        analysis_reaction_network_enrichment,
        analysis_ratio_network_enrichment,
        analysis_ratio_distribution,
        analysis_enrichment_ratio_plots,
        analysis_chain_length
    )
    # noinspection PyUnresolvedReferences
    from ._network_compute_statistics import (
        compute_correlations,
        compute_pvalues,
        compute_fold_changes,
        compute_correlation_changes,
        compute_partial_correlations,
        _correlation_computation_
    )
    # noinspection PyUnresolvedReferences
    from ._network_lipid_utils import (
        _get_weighted_lipid_dict,
        add_lipid,
        all_lipid_objects,
        all_native_lipids,
        get_incompatible_lipids,
        all_molecular_species_lipids
    )
    # noinspection PyUnresolvedReferences
    from ._network_lipidome_summary import (
        analysis_lipid_substructure,
        lipidome_summary,
        # TODO: add substructure selection including plotting
        substructure_selection
    )
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(
        self,
        data: Union[str, pd.DataFrame],
        fatty_acids: str = "",
        groups: Union[str, pd.Series] = None,
        lipids_are_columns: bool = True,
        node_attributes: List[str] = None,
        edge_attributes: List[str] = None,
        lipids_lm_compatible: bool = False,
        organism: str = "HSA",
        database: Union[str, List[str], Tuple[str]] = ("Rhea", "Reactome"),
        ether_conversions: bool = True,
        confirmed_lipid_species: str = "",
        allow_molspec_fails: bool = False,
        reference_group: str = None
    ):
        """

        Parameters
        ----------
        :param data : Union[str, pd.DataFrame]
            Path to a csv file that can be read with
            pandas.read_csv or pd.DataFrame

        :param fatty_acids: str, default=""
            Path to a txt file containing fatty acid
            settings in the form described on the
            LINEX homepage. If "", a default fatty acid file is used.

        :param groups: pandas.Series or str, optional
            Sample groups as a pandas.Series or a path to file
            which contains the sample groups.
            Series indices should correspond the columns of 'data'.

        :param lipids_are_columns: bool, default=True
            If True in the "data" table, lipids represent columns and samples, rows.
            If False, lipids are rows and samples columns.

        :param node_attributes: List[str], optional
            List of node attributes to be made available for plotting. Calling
            LipidNetwork.add_attributes_to_plot has the same effect. For a list
            of available options call linex2.node_colour_options

        :param edge_attributes: List[str], optional
            List of edge attributes to be made available for plotting. Calling
            LipidNetwork.add_attributes_to_plot has the same effect. For a list
            of available options call linex2.edge_colour_options

        :param lipids_lm_compatible: bool, default=False
            When the lipids in the data file are not compatible with the LIPID MAPS nomenclature,
            they can be converted with LipidLynxX. If the parameter is False,
            lipids will be converted before (This can take up to several minutes).
            If True, the lipids won't be converted.
            For more information visit the LipidLynxX website: https://www.lipidmaps.org/lipidlynxx/

        :param organism: str, default="HSA"
            Three letter organism code.
            We use reactions from the Reactome database, where an organism has to be selected.
            All available organisms can be found here: https://reactome.org/content/schema/objects/Species
            The default organism is Homo Sapiens. If the user provides an empty string "" all reactions from the
            database are considered.

        :param database:  Union[str, List[str], Tuple[str]], default = ("Rhea", "Reactome")
            Lipid reaction networks can be computed from the Reactome and Rhea database.
            Reactome is organism specific, Rhea does not provide this information.
            One or both databases can be selected and the network is
            computed using the reactions from those databases.

        :param ether_conversions: bool, default=True
            Include Conversions of Lipids to ether lipids in the network as L_FAmodify edges.

        :param reference_group: str, default=None
            Sets a specific group as a reference against which all pairwise metrics (e.g.
            p-values and fold-changes) are computed.
        """

        # Get reference lipids
        self._ref_dict = parse_lipid_reference_table_dict(STANDARD_LIPID_CLASSES)

        # Read data if necessary
        if isinstance(data, pd.DataFrame):
            user_data = data
        elif isinstance(data, str):
            user_data = pd.read_csv(data, index_col=0)
        else:
            raise ValueError("Data must be a str (path to data file) or pd.DataFrame")

        # Transpose, so that lipids are columns
        if not lipids_are_columns:
            user_data = user_data.T

        # Extract lipids
        data_lipids = user_data.columns.values
        if all([isinstance(lipid, int) for lipid in data_lipids]):
            raise ValueError(
                'Lipid axis names contain integers only. '
                'Did you forget to transpose your data or add the right column names?'
            )

        # Parse lipids
        lipid_dict, converted_names_mapping, incompatible_lipids = parse_lipid_list(data_lipids,
                                                                                    reference_lipids=self._ref_dict,
                                                                                    ll_compatible=lipids_lm_compatible)
        incompatible_2 = []
        lipid_dict_confirmed = dict()
        # Get confirmed species
        if confirmed_lipid_species != "":
            confirmed_species_file = open(confirmed_lipid_species, "r")
            confirmed_lines = [line.strip() for line in confirmed_species_file]
            lipid_dict_confirmed, _, incompatible_2 = parse_lipid_list(confirmed_lines,
                                                                       reference_lipids=self._ref_dict,
                                                                       ll_compatible=lipids_lm_compatible)

        # Check if enough lipids
        if len(incompatible_lipids) >= (len(data_lipids) - 2):
            raise InputDataError(f"Only {len(data_lipids) - len(incompatible_lipids)} lipids were identified "
                                 f"successfully. At least 3 lipids have to be identified to use LINEX2.")

        # Remove incompatible lipids
        user_data = user_data.drop(columns=incompatible_lipids + incompatible_2)

        self._incompatible_lipids = incompatible_lipids + incompatible_2  # save them for later access

        # Rename lipids with new names
        user_data = user_data.rename(columns=converted_names_mapping)
        # Aggregate lipids with the same name after lynx conversion
        if np.any(user_data.columns.duplicated()):
            user_data = user_data.T.groupby(level=0).sum().T

        # Uebergibt lipid_dict
        self._compute_gln_(
            lipid_dict=lipid_dict,
            fa_file=fatty_acids,
            organism=organism,
            database=database,
            ether_conversions=ether_conversions,
            confirmed_species_dict=lipid_dict_confirmed,
            allow_molspec_fails=allow_molspec_fails
        )

        self._set_data_groups_(user_data.T, groups, reference_group)  # Forward transposed user_data
        self.last_network_call = {}
        self.network = None

        # 'static' properties (i.e. data-indepentent)
        # these do NOT include graph metrics!
        self._attributes_to_plot = {
            "nodes": STATIC_NODE_PROPERTIES.union(self._allowed_node_attributes_),
            "edges": STATIC_EDGE_PROPERTIES.union(self._allowed_edge_attributes_)
        }
        self._reset_nodes_()
        self._reset_edges_()

        # adding the data-dependent attributes to plot
        self.add_attributes_to_plot(node_attributes, edge_attributes)

    def _compute_gln_(self, lipid_dict: Dict[str, List[Lipid]],
                      fa_file: str,
                      organism: str = "HSA",
                      database: Union[str, List[str], Tuple[str]] = ("Rhea", "Reactome"),
                      ether_conversions: bool = True,
                      confirmed_species_dict: Dict[str, List[Lipid]] = None,
                      allow_molspec_fails: bool = False
                      ):
        """
        ** Function for internal use only! **

        Build the GenerateLipidNetwork class with.

        :param lipid_dict:  Dict[str, List[Lipid]]
            Dict of parsed lipids from the data.
            Is computed e.g. from the function parse_lipid_list

        :param fa_file:
            Path to fatty acid file.

        :param organism: str, default="HSA"
            Three letter organism code.
            We use reactions from the Reactome database, where an organism has to be selected.
            All available organisms can be found here: https://reactome.org/content/schema/objects/Species
            The default organism is Homo Sapiens. If the user provides an empty string "" all reactions from the
            database are considered.

        :param database:  Union[str, List[str], Tuple[str]], default = ("Rhea", "Reactome")
            Lipid reaction networks can be computed from the Reactome and Rhea database.
            Reactome is organism specific, Rhea does not provide this information.
            One or both databases can be selected and the network is
            computed using the reactions from those databases.

        :return:
            Initializes slot: LipidNetwork.gln
        """
        if confirmed_species_dict is None:
            confirmed_species_dict = dict()

        ref_lips = parse_lipid_reference_table(STANDARD_LIPID_CLASSES)
        # del ref_dict["Cholesterol"]
        # del ref_dict["ST"]

        # Extract possible class reactions
        if isinstance(database, str):
            database = [database]

        if check_databases(database):  # Raises Error, if database not valid
            pass

        class_reaction_list = []
        if "Reactome" in database:
            if organism == "":
                class_reaction_list.append(make_all_reaction_list_from_reactome(REACTOME_OTHERS_UNIQUE,
                                                                                REACTOME_REACTION_CURATION,
                                                                                REACTOME_REACTION_TO_MOLECULE,
                                                                                self._ref_dict,
                                                                                REACTOME_REACTION_DETAILS,
                                                                                verbose=False)[0]
                                           )
            else:
                class_reaction_list.append(make_organism_reaction_list_from_reactome(REACTOME_OTHERS_UNIQUE,
                                                                                     REACTOME_REACTION_CURATION,
                                                                                     REACTOME_REACTION_TO_MOLECULE,
                                                                                     self._ref_dict,
                                                                                     REACTOME_REACTION_DETAILS,
                                                                                     verbose=False,
                                                                                     organism=organism)
                                           )

        if "Rhea" in database:
            class_reaction_list.append(make_all_reaction_list_from_rhea(RHEA_OTHERS_UNIQUE,
                                                                        RHEA_REACTION_CURATION,
                                                                        RHEA_REACTION_TO_MOLECULE,
                                                                        self._ref_dict,
                                                                        gn_mapping=RHEA_MAPPING,
                                                                        verbose=False)[0]
                                       )

        if len(class_reaction_list) == 1:
            # If only one database is used, the Reaction list can be used as is.
            class_reacs = class_reaction_list[0]
        else:
            # If both databases are used, they are combined.
            class_reacs = combine_reactions(class_reaction_list[0],
                                            class_reaction_list[1])

        # Parse fatty acid file
        if fa_file == "":
            with pkg_resources.path(data, "default_fatty_acids.txt") as path:
                fatty_acid_file = open(path, "r")
        else:
            fatty_acid_file = open(fa_file, "r")

        fa_lines = [line for line in fatty_acid_file]
        fa_reacs = parse_fa_reactions(fa_lines)
        # print(fa_reacs)
        ex_fa_reacs = parse_excluded_fa_reactions(fa_lines)
        class_fas = parse_fatty_acids(fa_lines)
        # Extend class fatty acids
        class_fas = extend_class_fa(class_fa=class_fas, lipid_dict=lipid_dict, add_to_all=True)
        self.gln = GenerateLipidNetwork(ref_lips, lipid_dict, class_reacs,
                                        fa_reacs, ex_fa_reacs, class_fas, ether_conversions=ether_conversions,
                                        confirmed_species_dict=confirmed_species_dict,
                                        allow_molspec_fails=allow_molspec_fails)

    def _update_attrs_to_plot_(self, attr: str, nodes: bool):
        type_ = "nodes" if nodes else "edges"
        allowed = self._allowed_node_attributes_ if nodes else self._allowed_edge_attributes_
        if attr in allowed:
            self._attributes_to_plot[type_].add(attr)

    def add_attributes_to_plot(
        self, node_attributes: List[str] = None,
        edge_attributes: List[str] = None
    ):
        if node_attributes is not None:
            for attr in self._attributes_to_plot["nodes"]:
                self._update_attrs_to_plot_(attr, True)
        if edge_attributes is not None:
            for attr in self._attributes_to_plot["edges"]:
                self._update_attrs_to_plot_(attr, False)

    def _reset_nodes_(self, reset_colours=True):
        # TODO: add colour only for the single added node => speed-up
        #       and expand this to other options once they are available
        if not hasattr(self, "lipid_attributes"):
            self.lipid_attributes = {}
        if self.network is not None:
            for attr in self._attributes_to_plot["nodes"]:
                if self._attr_group_size_[attr] == 0:
                    net_attr = nx.get_node_attributes(self.network, attr)
                    if net_attr:
                        self.lipid_attributes[attr] = net_attr
        if reset_colours:
            self.node_colours = {}

    def _reset_edges_(self, reset_colours=True):
        self.last_network_call = {}
        self.interaction_attributes = {
            attr: {} for attr in self._attributes_to_plot["edges"]
        }
        if reset_colours:
            self.edge_colours = {}

    # Excluded for now, since the data is necessary for parsing the lipids in the __init__ function.
    # def add_data(self, data: Union[pd.DataFrame, str]):
    #     self._set_data_groups_(data=data, groups=None)
    #
    # def add_groups(self, groups: Union[pd.Series, str]):
    #     self._set_data_groups_(data=None, groups=groups)

    def _check_data_(self, check_groups=True):
        if not hasattr(self, "data"):
            raise ValueError("p-values cannot be computed when no dataset is given. Please call 'add_data'")
        elif self.data is None:
            raise ValueError("p-values cannot be computed when no dataset is given. Please call 'add_data'")
        if check_groups:
            if not hasattr(self, "groups"):
                raise ValueError("p-values cannot be computed when no groups are given. Please call 'add_groups'")
            elif self.groups is None:
                raise ValueError("p-values cannot be computed when no groups are given. Please call 'add_groups'")

    def _set_data_groups_(self, data: Union[pd.DataFrame, str] = None,
                          groups: Union[pd.Series, str] = None,
                          reference_group: str = None):
        # TODO: adapt once species strings are converted
        #       to Lipid objects internally to match names
        if data is not None:
            if hasattr(self, "data"):
                warn("'data' has already been set, skipping...")
            elif isinstance(data, pd.DataFrame):
                self.data = data
            elif isinstance(data, str):
                self._set_data_groups_(data=pd.read_csv(data, index_col=0))
            else:
                raise ValueError(
                    f"'data' must be str or pd.DataFrame, not {type(data).__name__}"
                )
            # NOTE: we force column names to be strings in order
            #       to avoid mismatching due to different data types
            self.data.columns = self.data.columns.astype(str)
        if groups is not None:
            if hasattr(self, "groups") and self.groups is not None:
                warn("'groups' has already been set, skipping...")
            elif isinstance(groups, pd.Series):
                self.groups = groups
            elif isinstance(groups, str):
                self.groups = pd.Series(pd.read_csv(groups, index_col=0))
                if hasattr(self, "data"):
                    matched = np.isin(self.data.columns, self.groups.index)
                    if not matched.all():
                        raise InputDataError(
                            f"Samples '{self.data.columns[~matched]}' were not found in group annotation"
                        )
            else:
                raise ValueError(
                    f"'groups' must be str or pd.Series, not {type(groups).__name__}"
                )
            # NOTE: we force the index to be strings in order
            #       to avoid mismatching due to different data types
            self.groups.index = self.groups.index.astype(str)
            self.groups = self.groups.astype(str)
            try:
                self.groups = self.groups[self.data.columns]
            except KeyError:
                matched = np.isin(self.data.columns, self.groups.index)
                raise InputDataError(
                    f"Samples '{self.data.columns[~matched]}' were not found in group annotation"
                )
            # unique groups
            self.unique_groups = self.groups.unique()
            # all pairwise combinations of unique groups
            if reference_group is not None:
                if reference_group not in self.unique_groups:
                    warn(f"{reference_group} is not one of the given groups ({list(self.unique_groups)})")
                    self.comparisons = list(combinations(self.unique_groups, 2))
                else:
                    self.comparisons = [(reference_group, group) for group in self.unique_groups
                                        if group != reference_group]
                    self.reference_group = reference_group
            else:
                self.comparisons = list(combinations(self.unique_groups, 2))

        elif not hasattr(self, "groups"):
            self.groups = None

    def _compute_network_(
        self, network: str,
        filter_duplicates: bool = False,
        excluded_reaction_types: List[str] = None,
        bipartite: bool = False,
        bipartite_type: str = "enzyme",
        verbose: bool = False,
        force_overwrite: bool = False
    ):
        # TODO: rewrite this to adapt to new layout
        function_call = locals()
        _ = function_call.pop('self')
        if function_call == self.last_network_call and not force_overwrite:
            if verbose:
                warnings.warn(f"Current call '{function_call}' is equivalent to existing network.")
            return
        else:
            self.last_network_call = function_call
            if verbose:
                print(f"Function call: {function_call}")
        if network == "native":
            self.network, self.lipid_mapping = self.gln.native_network(
                filter_duplicates=filter_duplicates,
                excluded_reaction_types=excluded_reaction_types,
                return_lipid_mapping=True,
                bipartite=bipartite,
                bipartite_type=bipartite_type
            )
        else:
            self.network, self.lipid_mapping = self.gln.molecular_species_network(
                filter_duplicates=filter_duplicates,
                excluded_reaction_types=excluded_reaction_types,
                return_lipid_mapping=True,
                bipartite=bipartite,
                bipartite_type=bipartite_type
            )
        self.reverse_mapping = {
            net_name: data_name for data_name, net_names in self.lipid_mapping.items()
            for net_name in net_names
        }
        # NOTE: _reste_nodes guarantue that we compute metrics after
        #       last network call
        self._reset_nodes_(False)
        self.lipid_attributes["degree"] = dict(self.network.degree())
        self.lipid_attributes["betweenness"] = nx.betweenness_centrality(self.network)
        self.lipid_attributes["closeness"] = nx.closeness_centrality(self.network)
        # setting data independent attributes
        for node_attr in self._attributes_to_plot["nodes"]:
            if self._attr_group_size_[node_attr] == 0:
                self._generate_node_colours(
                    node_attr, self._discrete_map_[node_attr]
                )

        # reaction types and enzymes
        for attr in ["reaction_type", "enzyme_id"]:
            rts = pd.DataFrame(index=list(self.network.nodes), columns=list(self.network.nodes))
            for edge, rt in nx.get_edge_attributes(self.network, attr).items():
                rts.loc[edge[0], edge[1]] = rt
                rts.loc[edge[1], edge[0]] = rt
            self.interaction_attributes[attr] = SymmetricMatrix(rts)

    def compute_native_network(
        self, filter_duplicates: bool = False,
        excluded_reaction_types: List[str] = None,
        bipartite: bool = False,
        bipartite_type: str = "enzyme",
        **kwargs
    ):
        # TODO: adapt documentation
        """
        Computing the native (i.e. mixed species level) network

        Parameters
        ----------
        :param filter_duplicates : bool, optional, default True
            Whether duplicate edges should be removed.

        :param excluded_reaction_types : list, optional, default None
            Optional list of reaction types, that should be ignored when
            computing the network edges.
        """
        self._compute_network_(
            "native", filter_duplicates=filter_duplicates,
            excluded_reaction_types=excluded_reaction_types,
            bipartite=bipartite,
            bipartite_type=bipartite_type,
            **kwargs
        )

    def compute_molecular_network(
        self, filter_duplicates: bool = False,
        excluded_reaction_types: List[str] = None,
        bipartite: bool = False,
        bipartite_type: str = "enzyme",
        **kwargs
    ):
        # TODO: adapt documentation
        """
        Computing the molecular species network

        Parameters
        ----------
        :param filter_duplicates : bool, optional, default True
            Whether duplicate edges should be removed

        :param excluded_reaction_types : list, optional, default None
            Optional list of reaction types, that should be ignored when
            computing the network edges
        """
        self._compute_network_(
            "molecular", filter_duplicates=filter_duplicates,
            excluded_reaction_types=excluded_reaction_types,
            bipartite=bipartite,
            bipartite_type=bipartite_type,
            **kwargs
        )

    def add_network_visualisation(self):
        # TODO: add a check for bipartite
        for attr in self._attributes_to_plot["nodes"]:
            self._add_network_attribute_(attr)
        for attr in self._attributes_to_plot["edges"]:
            self._add_network_attribute_(attr, False)

    def _check_attributes_(self,
                           edge_colour_attr: Union[str, None],
                           node_colour_attr: Union[str, None]):
        """
        Auxiliary Function to check if edge and node colouring attributes have been
        computed previously

        Parameters
        ----------
        :param edge_colour_attr : str
            String indicating the attribute used for edge colouring.

        :param node_colour_attr :
            String indicating the attribute used for node colouring.
        """
        edge_colour_attr_ = {
            "correlations": "compute_correlations",
            "partial_correlations": "compute_partial_correlations",
            "correlation_changes": "compute_correlation_changes",
            "partial_correlation_changes": "compute_correlation_changes",
            "reaction_type": "",
        }
        node_colour_attr_ = {
            "fold_changes": "compute_fold_changes",
            "pvalues": "compute_pvalues",
            "nlog_pvalues": "compute_pvalues",
            "lipid_class": None,
            "c_index": None,
            "db_index": None,
            "oh_index": None,
            "chain_length": None,
            "desaturation": None,
            "hydroxylation": None,
            "closeness": None,
            "betweenness": None,
            "degree": None
        }

        not_in = "'{0}' must be one of {1}, not {2}"

        if edge_colour_attr is not None:
            if (edge_colour_attr not in edge_colour_attr_.keys() and
                edge_colour_attr not in self._allowed_edge_attributes_):
                raise ValueError(
                    not_in.format("edge_colour_attr", list(edge_colour_attr_.keys()),
                                  edge_colour_attr)
                )

            if edge_colour_attr != "reaction_type":
                if self.interaction_attributes is not None:
                    if (edge_colour_attr not in self.interaction_attributes.keys() and
                        not nx.get_edge_attributes(self.network, edge_colour_attr)):
                        raise NotComputedError(edge_colour_attr, edge_colour_attr_[edge_colour_attr])
                    elif (not self.interaction_attributes.get(edge_colour_attr) and
                          not nx.get_edge_attributes(self.network, edge_colour_attr)):
                        raise NotComputedError(edge_colour_attr, edge_colour_attr_[edge_colour_attr])
                else:
                    raise NotComputedError(edge_colour_attr, edge_colour_attr_[edge_colour_attr])
        if node_colour_attr is not None:
            if (node_colour_attr not in node_colour_attr_.keys() and
                node_colour_attr not in self._allowed_node_attributes_):
                raise ValueError(
                    not_in.format("node_colour_attr", list(node_colour_attr_.keys()),
                                  node_colour_attr)
                )
            if self.lipid_attributes is not None:
                if (not self.lipid_attributes.get(node_colour_attr) and
                    not nx.get_edge_attributes(self.network, edge_colour_attr)):
                    raise NotComputedError(node_colour_attr, node_colour_attr_[node_colour_attr])
            else:
                raise NotComputedError(node_colour_attr, node_colour_attr_[node_colour_attr])

    def _add_network_attribute_(
        self, attr: str, nodes: bool = True,
        group_subset: Union[tuple, list, set] = None,
        overwrite: bool = True, add_group: bool = True,
        from_group: bool = True
    ):
        if nodes:
            if not self.lipid_attributes.get(attr, {}):
                raise NotComputedError(
                    f"{attr} has not been found in the list of computed "
                    "lipid properties. Please call the appropriate function first!"
                )
        else:
            if not self.interaction_attributes.get(attr):
                raise NotComputedError(
                    f"{attr} has not been found in the list of computed "
                    "interaction properties. Please call the appropriate function first!"
                )

        if self.network is None:
            raise ValueError(
                "No network has been computed yet! Please do so and call this function again."
            )
        if not overwrite:
            if nodes:
                old = nx.get_node_attributes(self.network, attr)
            else:
                old = nx.get_edge_attributes(self.network, attr)
            if old is not None:
                if not isinstance(old, dict):
                    return
                # some attributes are dicts => need to check whether
                # they are empty
                if old:
                    return

        attr_size = self._attr_group_size_[attr]
        if self.groups is not None:
            if attr_size > 0:
                # checking if correct number of groups are provided
                # if no groups are provided but the number of unique
                # groups in self are the same as the number of groups
                # required for the respective attribute, they are used
                if group_subset is None:
                    if self.unique_groups.size == attr_size:
                        group_subset = self.unique_groups
                    else:
                        raise ValueError(
                            f"'{attr}' requires exactly {attr_size} groups, "
                            f"but {self.unique_groups.size} are specified in the class instance. "
                            "Please use 'group_subset' to specify a subset of groups."
                        )
                elif len(group_subset) != attr_size:
                    raise ValueError(
                        f"'{attr}' requires exactly {attr_size} groups, "
                        f"but {len(group_subset)} are given."
                    )
                # actual computations with the correct number of groups
                if len(group_subset) == 1:
                    group = group_subset[0]
                    if add_group:
                        attr_ = f"{attr}_{group}"
                    else:
                        attr_ = attr
                    if nodes:
                        if not nx.get_node_attributes(self.network, attr_):
                            for node in self.network.nodes:
                                self.network.nodes[node][attr_] = {}
                        for node in self.network.nodes:
                            node_name = node if attr in STATIC_NODE_PROPERTIES else self.reverse_mapping.get(node)
                            if add_group:
                                to_add = self.lipid_attributes[attr][group].get(node_name)
                            else:
                                to_add = self.lipid_attributes[attr].get(node_name)
                            if to_add is None and attr not in STATIC_NODE_PROPERTIES:
                                if node_name in self.data.index:
                                    # in this case the attribute should have been computed
                                    raise ValueError(
                                        f"Something went wrong while computing {attr}. "
                                        f"{node} is present in input data frame, but not in attribute values"
                                    )
                                else:
                                    # this means the node was added during network generation
                                    # and does not have any associated values
                                    # TODO: adapt this behaviour when parsing is all set up!
                                    to_add = np.nan
                            if hasattr(to_add, "item"):
                                # json.serialize cannot handle numpy data types
                                to_add = to_add.item()
                            if add_group:
                                self.network.nodes[node][attr_] = to_add
                            else:
                                self.network.nodes[node][attr][group] = to_add
                    else:
                        if not nx.get_edge_attributes(self.network, attr_):
                            for edge in self.network.edges:
                                self.network.edges[edge][attr_] = {}
                        for edge in self.network.edges:
                            if attr not in STATIC_EDGE_PROPERTIES:
                                edge_name = (self.reverse_mapping.get(edge[0]),
                                             self.reverse_mapping.get(edge[1]))
                            else:
                                edge_name = edge
                            if from_group:
                                to_add = self.interaction_attributes[attr][group].get(edge_name[0], edge_name[1])
                            else:
                                to_add = self.interaction_attributes[attr].get(edge_name[0], edge_name[1])
                            if to_add is None and attr not in STATIC_EDGE_PROPERTIES:
                                # TODO: adapt when parsing is all set up!
                                if edge_name[0] in self.data.index and edge_name[1] in self.data.index:
                                    raise ValueError(
                                        f"Something went wrong while computing {attr}. "
                                        f"{edge[0]} and {edge[1]} are present in input data frame, but not in attribute values"
                                    )
                                else:
                                    to_add = np.nan
                            if hasattr(to_add, "item"):
                                # json.serialize cannot handle numpy data types
                                to_add = to_add.item()
                            if add_group:
                                self.network.edges[edge][attr_] = to_add
                            else:
                                self.network.edges[edge][attr][group] = to_add
                else:
                    # NOTE: only 1 or 2 possible (no more groups allowed)
                    # => would work for more than 2 groups as well!
                    attr_computed = False
                    for perm in permutations(group_subset):
                        if nodes:
                            if self.lipid_attributes[attr].get(perm) is not None:
                                attr_computed = True
                                break
                        else:
                            if self.interaction_attributes[attr].get(perm) is not None:
                                attr_computed = True
                                break
                    if not attr_computed:
                        raise NotComputedError(attr, group_subset, for_subset=True)
                    group = tuple(group_subset)
                    if group is None:
                        raise ValueError(
                            "Groups in 'group_subset' are not matching the group "
                            "group names provided in the class instance."
                        )
                    if add_group:
                        attr_ = f"{attr}_{_tuple_to_string_(group)}"
                    else:
                        attr_ = attr
                    if nodes:
                        nx.set_node_attributes(self.network,
                                               self.gln.n_lipids * [""],
                                               name=attr_)
                        for node in self.network.nodes:
                            node_name = node if attr in STATIC_NODE_PROPERTIES else self.reverse_mapping.get(node)
                            if from_group:
                                to_add = self.lipid_attributes[attr][group].get(node_name)
                            else:
                                to_add = self.lipid_attributes[attr].get(node_name)
                            if to_add is None and attr not in STATIC_NODE_PROPERTIES:
                                if node_name in self.data.index:
                                    # in this case the attribute should have been computed
                                    raise ValueError(
                                        f"Something went wrong while computing {attr}. "
                                        f"{node} is present in input data frame, but not in attribute values"
                                    )
                                else:
                                    # this means the node was added during network generation
                                    # and does not have any associated values
                                    # TODO: adapt this behaviour when parsing is all set up!
                                    to_add = np.nan
                            if hasattr(to_add, "item"):
                                # json.serialize cannot handle numpy data types
                                to_add = to_add.item()
                            self.network.nodes[node][attr_] = to_add
                    else:
                        nx.set_edge_attributes(self.network,
                                               len(self.network.edges) * [""],
                                               name=attr_)
                        for edge in self.network.edges:
                            if attr not in STATIC_EDGE_PROPERTIES:
                                edge_name = (self.reverse_mapping.get(edge[0]),
                                             self.reverse_mapping.get(edge[1]))
                            else:
                                edge_name = edge
                            if from_group:
                                to_add = self.interaction_attributes[attr][group].get(edge_name[0], edge_name[1])
                            else:
                                to_add = self.interaction_attributes[attr].get(edge_name[0], edge_name[1])
                            if to_add is None and attr not in STATIC_EDGE_PROPERTIES:
                                # TODO: adapt when parsing is all set up!
                                if edge_name[0] in self.data.index and edge_name[1] in self.data.index:
                                    raise ValueError(
                                        f"Something went wrong while computing {attr}. "
                                        f"{edge[0]} and {edge[1]} are present in input data frame, but not in attribute values"
                                    )
                                else:
                                    to_add = np.nan
                            if hasattr(to_add, "item"):
                                # json.serialize cannot handle numpy data types
                                to_add = to_add.item()
                            self.network.edges[edge][attr_] = to_add
            else:
                if nodes:
                    for node in self.network.nodes:
                        node_name = node if attr in STATIC_NODE_PROPERTIES else self.reverse_mapping.get(node)
                        to_add = self.lipid_attributes[attr].get(node_name)
                        if to_add is None and attr not in STATIC_NODE_PROPERTIES:
                            if node_name in self.data.index:
                                # in this case the attribute should have been computed
                                raise ValueError(
                                    f"Something went wrong while computing {attr}. "
                                    f"{node} is present in input data frame, but not in attribute values"
                                )
                            else:
                                # this means the node was added during network generation
                                # and does not have any associated values
                                # TODO: adapt this behaviour when parsing is all set up!
                                to_add = np.nan
                        if hasattr(to_add, "item"):
                            # json.serialize cannot handle numpy data types
                            to_add = to_add.item()
                        self.network.nodes[node][attr] = to_add
                else:
                    for edge in self.network.edges:
                        if attr not in STATIC_EDGE_PROPERTIES:
                            edge_name = (self.reverse_mapping.get(edge[0]),
                                         self.reverse_mapping.get(edge[1]))
                        else:
                            edge_name = edge
                        to_add = self.interaction_attributes[attr].get(edge[0], edge[1])
                        if to_add is None and attr not in STATIC_EDGE_PROPERTIES:
                            # TODO: adapt when parsing is all set up!
                            if edge_name[0] in self.data.index and edge_name[1] in self.data.index:
                                raise ValueError(
                                    f"Something went wrong while computing {attr}. "
                                    f"{edge[0]} and {edge[1]} are present in input data frame, but not in attribute values"
                                )
                            else:
                                to_add = np.nan
                        if hasattr(to_add, "item"):
                            # json.serialize cannot handle numpy data types
                            to_add = to_add.item()
                        self.network.edges[edge][attr] = to_add
        else:
            if nodes:
                for node in self.network.nodes:
                    node_name = node if attr in STATIC_NODE_PROPERTIES else self.reverse_mapping.get(node)
                    to_add = self.lipid_attributes[attr].get(node_name)
                    if to_add is None and attr not in STATIC_NODE_PROPERTIES:
                        if node_name in self.data.index:
                            # in this case the attribute should have been computed
                            raise ValueError(
                                f"Something went wrong while computing {attr}. "
                                f"{node} is present in input data frame, but not in attribute values"
                            )
                        else:
                            # this means the node was added during network generation
                            # and does not have any associated values
                            # TODO: adapt this behaviour when parsing is all set up!
                            to_add = np.nan
                    if hasattr(to_add, "item"):
                        # json.serialize cannot handle numpy data types
                        to_add = to_add.item()
                    self.network.nodes[node][attr] = to_add
            else:
                for edge in self.network.edges:
                    if attr not in STATIC_EDGE_PROPERTIES:
                        edge_name = (self.reverse_mapping.get(edge[0]),
                                     self.reverse_mapping.get(edge[1]))
                    else:
                        edge_name = edge
                    to_add = self.interaction_attributes[attr].get(edge_name[0], edge_name[1])
                    if to_add is None and attr not in STATIC_EDGE_PROPERTIES:
                        # TODO: adapt when parsing is all set up!
                        if edge_name[0] in self.data.index and edge_name[1] in self.data.index:
                            raise ValueError(
                                f"Something went wrong while computing {attr}. "
                                f"{edge[0]} and {edge[1]} are present in input data frame, but not in attribute values"
                            )
                        else:
                            to_add = np.nan
                    if hasattr(to_add, "item"):
                        # json.serialize cannot handle numpy data types
                        to_add = to_add.item()
                    if attr == "enzyme_id" and not self.directed:
                        to_add += f"<br>{self.network.edges[(edge[1], edge[0])][attr]}"
                    self.network.edges[edge][attr] = to_add

    @staticmethod
    def _set_min_max_scale_(
        values: np.ndarray,
        vmax: float = None,
        min_to_zero: bool = True
    ) -> Tuple[float, float]:
        min_ = values.min()
        if min_ < 0:
            if vmax is None:
                max_ = max(abs(values.min()),
                           abs(values.max()))
            else:
                max_ = vmax
            min_ = -max_
        else:
            if min_to_zero:
                min_ = 0
            if vmax is None:
                max_ = values.max()
            else:
                max_ = vmax
        return min_, max_

    def _generate_edge_colours_(
        self, attr: str,
        discrete: bool = True, cmap: str = "tab10",
        group: Union[str, tuple] = None,
        vmin: float = None, vmax: float = None,
        colours_to_hex: bool = False,
        group_attr: str = None
    ) -> Tuple[dict, list]:
        if attr == "correlation_changes" or attr == "partial_correlation_changes":
            if colours_to_hex:
                edge_colour_map = {
                    "significant to unsignificant": to_hex("tab:cyan"),
                    "unsignificant to significant": to_hex("tab:green"),
                    "unchanged significant": to_hex("tab:red"),
                    "unsignificant": "#e1e5e7",
                    "positive to negative": to_hex("tab:blue"),
                    "negative to positive": to_hex("tab:orange"),
                    None: to_hex("black")
                }
            else:
                edge_colour_map = {
                    "significant to unsignificant": "tab:cyan",
                    "unsignificant to significant": "tab:green",
                    "unchanged significant": "tab:red",
                    "unsignificant": "#e1e5e7",
                    "positive to negative": "tab:blue",
                    "negative to positive": "tab:orange",
                    None: to_hex("black")
                }
            edge_colours = [
                edge_colour_map[self.interaction_attributes[attr][group].get(self.reverse_mapping.get(edge[0]),
                                                                             self.reverse_mapping.get(edge[1]))]
                for edge in self.network.edges
            ]
            return edge_colour_map, edge_colours
        if self._attr_group_size_[attr] == 0:
            group = None
        if discrete:
            if group is None:
                attr_set = self.interaction_attributes[attr]
                if isinstance(attr_set, dict):
                    keys = attr_set.keys()
                    if len(keys) == 1:
                        attr_set = attr_set[list(keys)[0]]
                    else:
                        raise ValueError(
                            f"Multiple group options found for {attr}."
                            " Please specify which should be used by using the 'groups' parameter"
                        )
                attr_val_set = unique_elements(attr_set)
            else:
                attr_val_set = unique_elements(self.interaction_attributes[attr][group])
            if attr_val_set.size > 10 and cmap == "tab10":
                cmap = "tab20"
            if attr_val_set.size > 20:
                custom_map = _generate_colormap_(attr_val_set.size)
                colours = [custom_map(i) for i in range(attr_val_set.size)]
            else:
                colours = [plt.get_cmap(cmap)(i)
                           for i in range(attr_val_set.size)]
            edge_colour_map = dict(zip(attr_val_set, colours))
            if group_attr is not None:
                attr = group_attr
            if colours_to_hex:
                edge_colours = [to_hex(edge_colour_map.get(self.network.edges[edge][attr], (0.84, 0.84, 0.84)))
                                for edge in self.network.edges]
            else:
                edge_colours = [edge_colour_map.get(self.network.edges[edge][attr], (0.84, 0.84, 0.84))
                                for edge in self.network.edges]
        else:
            if cmap == "tab10":
                # NOTE: seismic is bad here because 0 values will be non-visible
                cmap = "coolwarm"
            if group is None:
                vals = self.interaction_attributes[attr]
                if isinstance(vals, dict):
                    keys = vals.keys()
                    if len(keys) == 1:
                        vals = vals[list(keys)[0]]
                    else:
                        raise ValueError(
                            f"Multiple group options found for {attr}."
                            " Please specify which should be used by using the 'groups' parameter"
                        )
            else:
                vals = self.interaction_attributes[attr][group]
            val_arr = vals.data
            if np.isnan(val_arr).all():
                vals = 5

            min_, max_ = self._set_min_max_scale_(val_arr, vmax)
            if np.isnan(min_):
                min_ = val_arr[~np.isnan(val_arr)].min()
                if np.isnan(min_):
                    min_ = 0
            if np.isnan(max_) or min_ == max_:
                max_ = val_arr[~np.isnan(val_arr)].max()
                if np.isnan(max_):
                    max_ = min_ + 1

            scm = ScalarMappable(
                norm=Normalize(
                    vmin=min_,
                    vmax=max_
                ),
                cmap=cmap
            )
            if colours_to_hex:
                edge_colours = [to_hex(scm.to_rgba(vals.get(self.reverse_mapping.get(edge[0]),
                                                            self.reverse_mapping.get(edge[1]),
                                                            np.nan)))
                                for edge in self.network.edges]
            else:
                edge_colours = [scm.to_rgba(vals.get(self.reverse_mapping.get(edge[0]),
                                                     self.reverse_mapping.get(edge[1]),
                                                     np.nan))
                                for edge in self.network.edges]
            edge_colour_map = {"min": -max_,
                               "max": max_,
                               "cmap": cmap}

        self.edge_colours[attr] = (edge_colour_map, edge_colours)
        return edge_colour_map, edge_colours

    def _generate_node_colours(
        self, attr: str,
        discrete: bool = True, cmap: str = "tab10",
        group: Union[str, tuple] = None,
        vmin: float = None, vmax: float = None,
        colours_to_hex: bool = False,
        group_attr: str = None
    ) -> Tuple[dict, list]:
        if attr == "lipid_class":
            if colours_to_hex:
                # TODO: test mapping of unknown lipid colours
                node_colours = [
                    LIPID_CLASS_COLOURS.get(self.network.nodes[node]["lipid_class"], "#000000")
                    for node in self.network.nodes
                ]
                # NOTE: this is necessary to also include lipid classes not covered by our default
                # colour scheme
                node_colour_map = {
                    lipid_class: LIPID_CLASS_COLOURS.get(lipid_class, "#000000")
                    for lipid_class in np.unique(list(self.lipid_attributes["lipid_class"].values()))
                }
            else:
                node_colours = [
                    to_rgba(LIPID_CLASS_COLOURS.get(self.network.nodes[node]["lipid_class"], "black"))
                    for node in self.network.nodes
                ]
                # NOTE: this is necessary to also include lipid classes not covered by our default
                # colour scheme
                node_colour_map = {
                    lipid_class: to_rgba(LIPID_CLASS_COLOURS.get(lipid_class, "black"))
                    for lipid_class in np.unique(list(self.lipid_attributes["lipid_class"].values()))
                }
        else:
            if self._attr_group_size_[attr] == 0:
                group = None
            if discrete:
                if group is None:
                    attr_val_set = unique_elements(self.lipid_attributes[attr])
                else:
                    attr_val_set = unique_elements(self.lipid_attributes[attr][group])
                if attr_val_set.size > 10 and cmap == "tab10":
                    cmap = "tab20"
                if attr_val_set.size > 20:
                    colours = _generate_colormap_(attr_val_set.size)
                else:
                    colours = [plt.get_cmap(cmap)(i)
                               for i in range(attr_val_set.size)]
                node_colour_map = dict(zip(attr_val_set, colours))
                if group_attr is not None:
                    attr = group_attr
                if colours_to_hex:
                    node_colours = [to_hex(node_colour_map[self.network.nodes[node][attr]])
                                    for node in self.network.nodes]
                else:
                    node_colours = [node_colour_map[self.network.nodes[node][attr]]
                                    for node in self.network.nodes]
            else:
                if cmap == "tab10":
                    cmap = "seismic"
                if group is None:
                    try:
                        vals = self.lipid_attributes[attr]
                    except KeyError:
                        raise KeyError(
                            f'{attr} has not been computed yet!'
                        )
                else:
                    try:
                        vals = self.lipid_attributes[attr][group]
                    except KeyError:
                        raise KeyError(
                            f'{attr} has not been computed for group {group} yet!'
                        )

                val_arr = np.array(list(vals.values()))

                index_attr = attr not in ["c_index", "db_index", "chain_length", "desaturation",
                                          "oh_index", "hydroxylation"]
                min_, max_ = self._set_min_max_scale_(val_arr, vmax,
                                                      min_to_zero=index_attr)
                if np.isnan(min_):
                    min_ = np.min(val_arr[~np.isnan(val_arr)])
                    if np.isnan(min_):
                        min_ = 0
                if np.isnan(max_) or min_ == max_:
                    max_ = np.max(val_arr[~np.isnan(val_arr)])
                    if np.isnan(max_):
                        max_ = min_ + 1
                if vmin is not None:
                    min_ = vmin
                scm = ScalarMappable(
                    norm=Normalize(
                        vmin=min_,
                        vmax=max_
                    ),
                    cmap=cmap
                )
                if colours_to_hex:
                    node_colours = [to_hex(scm.to_rgba(vals.get(self.reverse_mapping.get(node), np.nan)))
                                    for node in self.network.nodes]
                else:
                    node_colours = [scm.to_rgba(vals.get(self.reverse_mapping.get(node), np.nan))
                                    for node in self.network.nodes]
                if np.isnan(val_arr).any():
                    node_colour_map = {"min": min_,
                                       "max": max_,
                                       "nan": np.nan,
                                       "cmap": cmap}
                else:
                    node_colour_map = {"min": min_,
                                       "max": max_,
                                       "cmap": cmap}

        self.node_colours[attr] = (node_colour_map, node_colours)
        return node_colour_map, node_colours

    def _discrete_legend_(
        self, attr: str,
        nodes: bool = True, **kwargs
    ) -> List[Line2D]:
        if nodes:
            # node labels => scatters
            legend_map = self.node_colours[attr][0]
            markersize = kwargs.pop("markersize", 12)
            handles = [Line2D([0], [0], color="w", marker="o",
                              markerfacecolor=col, label=group,
                              markersize=markersize,
                              **kwargs)
                       for group, col in legend_map.items()]
        else:
            # edge labels => thick lines
            legend_map = self.edge_colours[attr][0]
            lw = kwargs.pop("lw", 4)
            handles = [Line2D([0], [0], color=col,
                              lw=lw, label=group,
                              **kwargs)
                       for group, col in legend_map.items()]

        return handles

    def _continuous_legend_(
        self, attr: str,
        ax: plt.axis = None,
        nodes: bool = True, **kwargs
    ) -> Colorbar:
        if nodes:
            legend_map = self.node_colours[attr][0]
        else:
            legend_map = self.edge_colours[attr][0]
        scm = ScalarMappable(norm=Normalize(vmin=legend_map["min"],
                                            vmax=legend_map["max"]),
                             cmap=legend_map["cmap"])
        return plt.colorbar(scm, ax=ax, **kwargs)

    @staticmethod
    def _scale_dict_(
        sizes: Dict[str, float],
        scale: Tuple[float, float],
        abs_: bool = False,
        map_as_ex: bool = False,
        vmax: float = None,
        map_data_names: dict = None
    ) -> Tuple[Dict[str, float], Dict[float, float]]:
        scale_sizes = np.array(list(sizes.values()))
        if np.isnan(scale_sizes).all():
            # arbitraty default
            scale_sizes = np.zeros(scale_sizes.shape) + 5
        # if abs_:
        #     scale_sizes = abs(scale_sizes)  # NOTE: in this case min = 0
        # else:
        nans = np.isnan(scale_sizes)
        if vmax is None:
            max_ = scale_sizes[~nans].max()
        else:
            max_ = vmax
        if abs_:
            min_ = 0
            scale_sizes = abs(scale_sizes)
        else:
            min_ = scale_sizes[~nans].min()
            scale_sizes = scale_sizes - min_
        if max_ - min_ != 0:
            scale_sizes = scale_sizes / (max_ - min_)
        if abs_:
            scale_sizes = abs(scale_sizes * (scale[1] - scale[0]) + scale[0])
        else:
            scale_sizes = scale_sizes * (scale[1] - scale[0]) + scale[0]
        if map_data_names:
            expanded_scaled = []
            lipid_names = []
            for idx, lipid in enumerate(sizes.keys()):
                for mapping in map_data_names.get(lipid, []):
                    expanded_scaled.append(float(scale_sizes[idx]))
                    lipid_names.append(mapping)
            scaled_sizes = dict(zip(lipid_names, expanded_scaled))
        else:
            native_type_list = [float(x) for x in scale_sizes]
            scaled_sizes = dict(zip(sizes.keys(),
                                    native_type_list))
        final_min = scale_sizes[~nans].min()
        final_max = scale_sizes[~nans].max()
        if map_as_ex:
            if nans.any():
                scale_map = {
                    "min": (float(min_), float(final_min)),
                    "max": (float(max_), float(final_max)),
                    # arbitrary values to draw nans in legend
                    "nan": (1, 1)
                }
            else:
                scale_map = {
                    "min": (float(min_), float(final_min)),
                    "max": (float(max_), float(final_max))
                }
        else:
            scale_sizes[nans] = final_min / 2
            scale_sizes = [float(x) for x in scale_sizes]
            scale_map = dict(zip(sizes.values(), scale_sizes))
        return scaled_sizes, scale_map

    def _edge_colour_subset_(
        self, attr: str,
        discrete: bool = True, cmap: str = "tab10",
        group: Union[str, tuple] = None,
        vmin: float = None, vmax: float = None
    ) -> Tuple[dict, list]:
        if group is None:
            return self._generate_edge_colours_(
                attr,
                discrete, cmap,
                vmin=vmin, vmax=vmax
            )
        elif len(group) == 1:
            return self._generate_edge_colours_(
                attr,
                discrete, cmap,
                group[0], vmax=vmax,
                vmin=vmin
            )
        else:
            return self._generate_edge_colours_(
                attr,
                discrete, cmap,
                group, vmax=vmax,
                vmin=vmin
            )

    def _node_colour_subset_(
        self, attr: str,
        discrete: bool = True, cmap: str = "tab10",
        group: Union[str, tuple] = None,
        vmin: float = None, vmax: float = None
    ) -> Tuple[dict, list]:
        if group is None:
            return self._generate_node_colours(
                attr,
                discrete, cmap,
                vmax=vmax, vmin=vmin)
        elif len(group) == 1:
            return self._generate_node_colours(
                attr,
                discrete, cmap,
                group[0], vmax=vmax,
                vmin=vmin
            )
        else:
            return self._generate_node_colours(
                attr,
                discrete, cmap,
                group, vmax=vmax,
                vmin=vmin
            )

    def plot_static_network(
        self,
        layout: Union[Callable, dict] = nx_layout.fruchterman_reingold_layout,
        overwrite_layout: bool = False,
        as_undirected: bool = True,
        edge_colour_attr: str = None,
        overwrite_edge_attr: bool = True,
        edge_map: str = "tab10",
        edge_legend: bool = True,
        edge_size_func: Union[Callable, int] = None,
        edge_size_scale: Tuple[int, int] = (.2, .3),
        edge_size_legend: bool = True,
        edge_size_title: str = "Edge size",
        node_colour_attr: str = None,
        overwrite_node_attr: bool = True,
        node_map: str = "tab10",
        node_legend: bool = True,
        node_size_func: Union[Callable, int] = None,
        node_size_scale: Tuple[int, int] = (50, 400),
        node_size_legend: bool = True,
        node_size_title: str = "Node size",
        edge_group_subset: Union[tuple, list, set] = None,
        node_group_subset: Union[tuple, list, set] = None,
        vmin: float = None,
        vmax: float = None,
        ax: plt.axis = None,
        figargs: dict = None,
        layout_args: dict = None,
        show: bool = False,
        **kwargs
    ) -> plt.axis:
        """
        Plotting the computed network in a static manner using networkx and matplotlib

        Parameters
        ----------
        :param layout: networkx.layout function or dict, optional, default nx_layout.fruchterman_reingold_layout
            If a dict specifying node position this is used to
            plot the network. If a callable it is used to
            compute the layout.
            If self.layout is None or overwrite_layout is True
            it will be stored in self.layout

        :param overwrite_layout: bool, optional, default False
            Whether to overwrite self.layout

        :param as_undirected: bool, optional, default True
            This parameter currently has no effect!
            Whether to show the network as directed or undirected.
            If directed, lipid class reactions will not be symmetric,
            i.e. PC -> PE could be possible but not PE -> PC.

        :param edge_colour_attr: str, optional
            Attribute to use for edge colours.
            Available options: correlations, partial_correlations,
            reaction_type, correlation_changes, partial_correlation_changes.
            All metrics except reaction_type have to be computed beforehand.

        :param overwrite_edge_attr: bool, optional, default True
            Whether to overwrite attribute with the given group setting in
            self.interaction_attributes

        :param edge_map: str, optional, default 'tab10'
            Colourmap to use for edge colouring. Default for continuous
            attributes is 'viridis'

        :param edge_legend: bool, optional, default True
            Whether to show a legend for edge colours

        :param edge_size_func: Callable or int, optional
            Used to compute edge sizes if a function.
            Else set as constant edge size.

        :param edge_size_scale: tuple, optional, default (3, 12)
            Tuple of integers specifying the minimum and maximum
            of plotted edge sizes. The actual values will be scaled
            to fit this range.

        :param edge_size_legend: bool, optional, True
            Whether to show the legend for edge sizes

        :param edge_size_title: str, optional, default 'Edge size'
            Title of the edge size legend

        :param node_colour_attr: str, optional
            Specifies the attribute to use for determining node colours.
            Available options are: fold_changes, pvalues, nlog_pvalues,
            lipid_class, c_index, db_index

        :param overwrite_node_attr: bool, optional, default True
            Whether to overwrite attribute with the given group setting in
            self.lipid_attributes

        :param node_map: str, optional, default 'tab10'
            Colourmap to use for node colouring. Default for continuous
            attributes is 'viridis'

        :param node_legend: bool, optional, default True
            Whether to show node colour legend

        :param node_size_func: Callable or int, optional
            Used to compute node sizes if a function.
            Else set as constant node size.

        :param node_size_scale: tuple, optional, default (3, 12)
            Tuple of integers specifying the minimum and maximum
            of plotted node sizes. The actual values will be scaled
            to fit this range.

        :param node_size_legend: bool, optional, default True
            Whether to show the legend for node sizes

        :param node_size_title: str, optional, default 'Node size'
            Title of node size legend

        :param edge_group_subset: tuple or list of size 2, optional
            Group-comparison to show in edge colouring/sizing

        :param node_group_subset: tuple or list of size 2, optional
            Group-comparison to show in node colouring/sizing

        :param vmin: float, optional
            Minimum value of the legend (for continuous scales).
            If not given it will be set to 0 (for only positive
            arrays) or to -max(abs(min), abs(max)) (for zero-centred
            colours).
            Setting vmin can lead to un-centred colour scales!

        :param vmax: float, optional
            Maximum value of the legend (for continuous scales).
            If vmin is not specified, the minimum value will be
            either 0 (if all values are positive) or -vmax

        :param ax: plt.axis, optional
            matplotlib axis to plot onto

        :param figargs: dict, optional
            keyword arguments to pass to plt.figure.
            Only relevant if ax=None (default)

        :param layout_args: dict, optional
            keyword arguments to pass to layout.
            Only relevant if layout is not a function (default)

        :param show : bool, optional, default False
            Whether to show the plot

        :param kwargs:
            keyword arguments to pass to networkx.draw


        Returns
        -------
        :return plt.axis on which the network is plotted


        Raises
        ------
        NotComputedError
            If edge_colour_attr or node_colour_attr have not been
            computed before

        ValueError
            If edge_colour_attr or node_colour_attr are not in the
            list of available options
        """
        edge_discrete = None
        node_discrete = None
        if edge_colour_attr is not None:
            self._check_attributes_(
                edge_colour_attr,
                node_colour_attr
            )
            # check if attributes have been added to network
            # edge attribute
            self._add_network_attribute_(
                edge_colour_attr, nodes=False,
                group_subset=edge_group_subset,
                overwrite=overwrite_edge_attr
            )
            edge_discrete = self._discrete_map_[edge_colour_attr]
            edge_colours = self._edge_colour_subset_(edge_colour_attr,
                                                     edge_discrete,
                                                     edge_map,
                                                     group=edge_group_subset)
        else:
            edge_legend = False
            edge_colours = ["tab:gray" for _ in range(len(self.network.edges))]
        # node attribute
        if node_colour_attr is not None:
            self._add_network_attribute_(
                node_colour_attr, nodes=True,
                group_subset=node_group_subset,
                overwrite=overwrite_node_attr
            )
            node_discrete = self._discrete_map_[node_colour_attr]
            node_colours = self._node_colour_subset_(node_colour_attr,
                                                     node_discrete,
                                                     node_map,
                                                     node_group_subset,
                                                     vmin=vmin, vmax=vmax)
        else:
            node_legend = False
            node_colours = ["tab:blue" for _ in range(len(self.network.nodes))]

        # node sizes:
        # ===========
        node_size_legend_ = False
        node_sizes = None
        if kwargs.get("node_size") is not None:
            node_size = kwargs.pop("node_size")
            if isinstance(node_size, int) or isinstance(node_size, float):
                node_size = {node: node_size for node in self.network.nodes}
            elif isinstance(node_size, list) or isinstance(node_size, set):
                node_size = dict(zip(self.network.nodes, node_size))
        else:
            if node_size_func is None:
                sizes = nx.betweenness_centrality(self.network)
                scale_sizes = _range_scale_(x=np.array(list(sizes.values())),
                                            a=node_size_scale[0],
                                            b=node_size_scale[1])
                node_size = dict(zip(sizes.keys(),
                                     list(scale_sizes)))
                node_size_legend_ = True
                node_sizes = (min(list(sizes.values())),
                              max(list(sizes.values())))
            elif isinstance(node_size_func, int):
                node_size = dict(zip(self.network.nodes,
                                     len(self.network.nodes) * [node_size_func]))
            elif isinstance(node_size_func, Callable):
                node_size = node_size_func(self.network)
                node_size_legend_ = True
            else:
                node_size = dict(zip(self.network.nodes,
                                     len(self.network.nodes) * [23]))

        # edge sizes:
        # ===========
        edge_sizes = None
        if kwargs.get("width") is not None:
            edge_size = kwargs.pop("width")
            edge_size_legend_ = False
        else:
            edge_size_legend_ = False
            if edge_size_func is None:
                sizes = nx.edge_betweenness_centrality(self.network)
                scale_sizes = _range_scale_(x=np.array(list(sizes.values())),
                                            a=edge_size_scale[0],
                                            b=edge_size_scale[1])
                edge_size = dict(zip(sizes.keys(),
                                     list(scale_sizes)))
                edge_size_legend_ = True
                edge_sizes = (min(list(sizes.values())),
                              max(list(sizes.values())))
            elif isinstance(edge_size_func, int):
                edge_size = dict(zip(self.network.edges,
                                     len(self.network.edges) * [edge_size_func]))
            elif isinstance(edge_size_func, Callable):
                edge_size = edge_size_func(self.network)
                if not isinstance(edge_size, dict):
                    edge_size = dict(zip(self.network.edges,
                                         edge_size))
                edge_size_legend_ = True
            else:
                edge_size = dict(zip(self.network.edges,
                                     len(self.network.edges) * [5]))

        if ax is None:
            if figargs is None:
                fig, ax = plt.subplots()
            else:
                fig, ax = plt.subplots(**figargs)
            ax.set_xticks([])
            ax.set_yticks([])

        pre_layout = hasattr(self, "layout")
        if not pre_layout:
            if layout_args is None:
                self.layout = layout(self.network)
            else:
                self.layout = layout(self.network, **layout_args)
        elif overwrite_layout or self.layout is None:
            if layout_args is None:
                self.layout = layout(self.network)
            else:
                self.layout = layout(self.network, **layout_args)

        if as_undirected:
            # NOTE: this has no influence here!
            if isinstance(self.network, nx.DiGraph):
                nx.draw(self.network.to_undirected(),
                        node_color=node_colours[1],
                        edge_color=edge_colours[1],
                        ax=ax,
                        node_size=[node_size[node] for node in self.network.nodes],
                        width=[edge_size[edge] for edge in self.network.edges],
                        pos=self.layout,
                        **kwargs)
            else:
                nx.draw(self.network,
                        node_color=node_colours[1],
                        edge_color=edge_colours[1],
                        ax=ax,
                        node_size=[node_size[node] for node in self.network.nodes],
                        width=[edge_size[edge] for edge in self.network.edges],
                        pos=self.layout,
                        **kwargs)
        else:
            nx.draw(self.network,
                    node_color=node_colours[1],
                    edge_color=edge_colours[1],
                    ax=ax,
                    node_size=[node_size[node] for node in self.network.nodes],
                    width=[edge_size[edge] for edge in self.network.edges],
                    pos=self.layout,
                    **kwargs)
        # colour legends
        # ==============
        if node_legend:
            if node_discrete:
                hands = self._discrete_legend_(node_colour_attr,
                                               nodes=True)
                if edge_discrete:
                    plt.gca().add_artist(
                        plt.legend(handles=hands,
                                   title=node_colour_attr,
                                   loc=1,
                                   bbox_to_anchor=(1.2, 1))
                    )
                else:
                    plt.gca().add_artist(
                        plt.legend(handles=hands,
                                   title=node_colour_attr,
                                   loc=1,
                                   bbox_to_anchor=(1.35, 1))
                    )
            else:
                nbar = self._continuous_legend_(node_colour_attr,
                                                ax, nodes=True)
                nbar.ax.set_title(node_colour_attr)
        if edge_legend:
            if edge_discrete:
                hands = self._discrete_legend_(edge_colour_attr,
                                               nodes=False)
                plt.gca().add_artist(
                    plt.legend(handles=hands,
                               title=edge_colour_attr,
                               loc=1,
                               bbox_to_anchor=(1.35, 1))
                )
            else:
                ebar = self._continuous_legend_(edge_colour_attr,
                                                ax, nodes=False)
                ebar.ax.set_title(edge_colour_attr)
        # size legends
        # ==============
        # TODO: choose node/edge size in a way such that the legend
        #       can represent the actual property
        if node_size_legend and node_size_legend_:
            if node_sizes is None:
                node_sizes = (min(list(node_size.values())),
                              max(list(node_size.values())))
                node_size_scale = None
            plt.gca().add_artist(
                plt.legend(handles=_size_legend_(node_sizes,
                                                 node_size_scale,
                                                 nodes=True),
                           loc="upper left", bbox_to_anchor=(-.15, 1),
                           title=node_size_title)
            )
        if edge_size_legend and edge_size_legend_:
            if edge_sizes is None:
                edge_sizes = (min(list(edge_size.values())),
                              max(list(edge_size.values())))
                edge_size_scale = None
            plt.gca().add_artist(
                plt.legend(handles=_size_legend_(edge_sizes,
                                                 edge_size_scale,
                                                 nodes=False),
                           loc="upper left", bbox_to_anchor=(-.15, .7),
                           title=edge_size_title)
            )

        plt.tight_layout()
        if show:
            plt.show()
        return ax

    def network_to_pyvis(
        self,
        edge_colour_attr: str = None,
        overwrite_edge_attr: bool = True,
        edge_cmap: str = "tab10",
        edge_size_func: Union[Callable, int] = None,
        edge_size_scale: Tuple[int, int] = (.2, .3),
        node_colour_attr: str = None,
        overwrite_node_attr: bool = True,
        node_cmap: str = "tab10",
        node_size_func: Union[Callable, int] = None,
        node_size_scale: Tuple[int, int] = (50, 400),
        edge_group_subset: Union[tuple, list, set] = None,
        node_group_subset: Union[tuple, list, set] = None,
        net: VisParser = None,
        as_directed: bool = False,
        vmin: float = None,
        vmax: float = None,
        **kwargs
    ) -> VisParser:
        """
        Converting the generated network from networkx format to pyvis format

        Parameters
        ----------
        :param edge_colour_attr: str, optional
            Attribute to use for edge colours.
            Available options: correlations, partial_correlations,
            reaction_type, correlation_changes, partial_correlation_changes.
            All metrics except reaction_type have to be computed beforehand.

        :param overwrite_edge_attr: bool, optional, default True
            Whether to overwrite attribute with the given group setting in
            self.interaction_attributes

        :param edge_cmap: str, optional, default 'tab10'
            Colourmap to use for edge colouring. Default for continuous
            attributes is 'viridis'

        :param edge_size_func: Callable or int, optional
            Used to compute edge sizes if a function.
            Else set as constant edge size.

        :param edge_size_scale: tuple, optional, default (3, 12)
            Tuple of integers specifying the minimum and maximum
            of plotted edge sizes. The actual values will be scaled
            to fit this range.

        :param node_colour_attr: str, optional
            Specifies the attribute to use for determining node colours.
            Available options are: fold_changes, pvalues, nlog_pvalues,
            lipid_class, c_index, db_index

        :param overwrite_node_attr: bool, optional, default True
            Whether to overwrite attribute with the given group setting in
            self.lipid_attributes

        :param node_size_func: Callable or int, optional
            Used to compute node sizes if a function.
            Else set as constant node size.

        :param node_size_scale: tuple, optional, default (3, 12)
            Tuple of integers specifying the minimum and maximum
            of plotted node sizes. The actual values will be scaled
            to fit this range.

        :param node_cmap: str, optional, default 'tab10'
            Colourmap to use for edge colouring. Default for continuous
            attributes is 'viridis'

        :param edge_group_subset: tuple or list of size 2, optional
            Group-comparison to show in edge colouring/sizing

        :param node_group_subset: tuple or list of size 2, optional
            Group-comparison to show in node colouring/sizing

        :param net: pyvis.network.Network, optional
            If given this instance will be used instead of generating
            a new one within the function

        :param as_directed: bool, optional, default False
            This has currenlty no effect!
            Whether to show a direted graph

        :param vmin: float, optional
                Minimum value of the legend (for continuous scales).
                If not given it will be set to 0 (for only positive
                arrays) or to -max(abs(min), abs(max)) (for zero-centred
                colours).
                Setting vmin can lead to un-centred colour scales!

        :param vmax : float, optional
            Maximum value of the legend (for continuous scales).
            If vmin is not specified, the minimum value will be
            either 0 (if all values are positive) or -vmax


        :param kwargs:
            keyword arguments to pass to pyvis.network.Network. Only
            releveant if net=None (default)

        Returns
        -------
        :return linex2.vis_utils.VisParser
        """
        self._check_attributes_(
            edge_colour_attr, node_colour_attr
        )
        # check if attributes have been added to network
        # edge attribute
        self._add_network_attribute_(
            edge_colour_attr, nodes=False,
            group_subset=edge_group_subset,
            overwrite=overwrite_edge_attr
        )
        edge_discrete = self._discrete_map_[edge_colour_attr]
        # node attribute
        self._add_network_attribute_(
            node_colour_attr, nodes=True,
            group_subset=node_group_subset,
            overwrite=overwrite_node_attr
        )
        node_discrete = self._discrete_map_[node_colour_attr]

        # generate colours
        edge_colours = self._edge_colour_subset_(edge_colour_attr,
                                                 edge_discrete,
                                                 edge_cmap,
                                                 edge_group_subset)
        node_colours = self._node_colour_subset_(node_colour_attr,
                                                 node_discrete,
                                                 node_cmap,
                                                 node_group_subset,
                                                 vmin=vmin, vmax=vmax)

        int_net = self.network.copy()
        if node_size_func is None:
            sizes = nx.betweenness_centrality(int_net)
            scale_sizes = _range_scale_(x=np.array(list(sizes.values())),
                                        a=node_size_scale[0],
                                        b=node_size_scale[1])
            node_size = dict(zip(sizes.keys(),
                                 list(scale_sizes)))
        elif isinstance(node_size_func, int):
            node_size = dict(zip(int_net.nodes,
                                 len(int_net.nodes) * [node_size_func]))
        elif isinstance(node_size_func, Callable):
            node_size = node_size_func(int_net)
        elif isinstance(node_size_func, dict):
            node_size = node_size_func
        else:
            node_size = dict(zip(int_net.nodes,
                                 len(int_net.nodes) * [5]))
        for i, node in enumerate(int_net.nodes):
            int_net.nodes[node]["value"] = node_size[node]
            int_net.nodes[node]["color"] = to_hex(node_colours[1][i])
            node_title = f"<strong>{node}</strong>:"
            prime = "<br><br><strong>{0}</strong>:<br>{1}"
            for prop in ["fold_changes", "pvalues"]:
                if self.lipid_attributes.get(prop) is not None:
                    sup = re.sub("s$", "", prop)
                    group_vals = self.lipid_attributes[prop]
                    # if fold_changes and pvalues are computes groups
                    # must have been specified at some point
                    for group, vals in group_vals.items():
                        node_title += prime.format(sup, f"{group}: {vals[node]}")
            int_net.nodes[node]["title"] = node_title

        if edge_size_func is None:
            sizes = nx.edge_betweenness_centrality(int_net)
            scale_sizes = _range_scale_(x=np.array(list(sizes.values())),
                                        a=edge_size_scale[0],
                                        b=edge_size_scale[1])
            edge_size = dict(zip(sizes.keys(),
                                 list(scale_sizes)))
        elif isinstance(edge_size_func, int):
            edge_size = dict(zip(int_net.edges,
                                 len(int_net.edges) * [edge_size_func]))
        elif isinstance(edge_size_func, Callable):
            edge_size = edge_size_func(int_net)
        elif isinstance(edge_size_func, dict):
            edge_size = edge_size_func
        else:
            edge_size = dict(zip(int_net.edges,
                                 len(int_net.edges) * [5]))

        props_oi = [prop for prop, val in self.interaction_attributes.items()
                    if val is not None]
        if isinstance(edge_group_subset, list):
            if len(edge_group_subset) == 1:
                def get_prop(x):
                    return self.interaction_attributes[x][edge_group_subset[0]]
            else:
                def get_prop(x):
                    return self.interaction_attributes[x][tuple(edge_group_subset)]
        else:
            def get_prop(x):
                return self.interaction_attributes[x]
        for i, edge in enumerate(int_net.edges):
            int_net.edges[edge]["value"] = edge_size[edge]
            int_net.edges[edge]["color"] = to_hex(edge_colours[1][i])
            edge_title = f"<strong>{int_net.edges[edge]['reaction_type']}</strong><br>"
            prime = "<strong>{0}</strong>:<br>{1}<br><br>"
            for prop in props_oi:
                sup = re.sub("s$", "", prop)
                if self._attr_group_size_[prop] == 0:
                    if prop != "enzyme_id":
                        val = self.interaction_attributes[prop].get(edge[0], edge[1], '')
                    else:
                        if as_directed:
                            renz = self.interaction_attributes[prop].get(edge[0], edge[1], '')
                            if isinstance(renz, str):
                                val = f"{renz.split(':')[1]}"
                            else:
                                val = str(renz)
                        else:
                            val_ij = self.interaction_attributes[prop].get(edge[0], edge[1], '')
                            val_ji = self.interaction_attributes[prop].get(edge[1], edge[0], '')
                            val = f"{val_ij}<br>{val_ji}"
                else:
                    try:
                        val = get_prop(prop).loc[edge[0], edge[1]]
                    # FIXME
                    # this is not a nice solution => make sure it's possible to
                    # correctly get edge values when supplying single groups
                    # for nodes
                    except KeyError:
                        val = None
                if val is not None:
                    if isinstance(val, float):
                        edge_title += prime.format(sup, round(val, ndigits=3))
                    else:
                        edge_title += prime.format(sup, val)
            int_net.edges[edge]["title"] = re.sub("<br>$", "", edge_title)

        if net is None:
            net = VisParser(directed=as_directed & isinstance(int_net, nx.DiGraph),
                            **kwargs)
        elif isinstance(net, Network):
            net = VisParser.from_pyvis_network(net, **kwargs)
        net.inherit_edge_colors(False)
        if as_directed and isinstance(int_net, nx.DiGraph):
            net.from_nx(int_net.to_directed())
        else:
            net.from_nx(int_net)
        # TODO: adapt to take continuous colour maps
        #       and node/edge sizes
        net.generate_legend(
            node_colours=node_colours[0],
            edge_colours=edge_colours[0]
        )
        return net

    def plot_interactive_network(
        self,
        file: str = "LipidNetwork.html",
        edge_colour_attr: str = None,
        overwrite_edge_attr: bool = True,
        edge_cmap: str = "tab10",
        edge_size_func: Union[Callable, int] = None,
        edge_size_scale: Tuple[int, int] = (2e-4, 3e-2),
        node_colour_attr: str = None,
        overwrite_node_attr: bool = True,
        node_cmap: str = "tab10",
        node_size_func: Union[Callable, int] = None,
        node_size_scale: Tuple[int, int] = (9, 11),
        edge_group_subset: Union[tuple, list, set] = None,
        node_group_subset: Union[tuple, list, set] = None,
        save: bool = False,
        show: bool = False,
        do_return: bool = False,
        notebook: bool = False,
        net: Network = None,
        show_filters: Union[str, List[str]] = None,
        as_directed: bool = False,
        vmin: float = None, vmax: float = None,
        **kwargs
    ) -> Union[VisParser, None]:
        """
        Plotting the computed network interactively using pyvis/vis.js

        Parameters
        ----------
        :param file: str, optional, default 'LipidNetwork.html'
            Path to which the html file (self-contained) is saved to.

        :param edge_colour_attr: str, optional
            Attribute to use for edge colours.
            Available options: correlations, partial_correlations,
            reaction_type, correlation_changes, partial_correlation_changes.
            All metricx except reaction_type have to be computed beforehand.

        :param overwrite_edge_attr: bool, optional, default True
            Whether to overwrite attribute with the given group setting in
            self.interaction_attributes

        :param edge_cmap: str, optional, default 'tab10'
            Colourmap to use for edge colouring. Default for continuous
            attributes is 'viridis'

        :param edge_size_func: Callable or int, optional
            Used to compute edge sizes if a function.
            Else set as constant edge size.

        :param edge_size_scale: tuple, optional, default (3, 12)
            Tuple of integers specifying the minimum and maximum
            of plotted edge sizes. The actual values will be scaled
            to fit this range.

        :param node_colour_attr: str, optional
            Specifies the attribute to use for determining node colours.
            Available options are: fold_changes, pvalues, nlog_pvalues,
            lipid_class, c_index, db_index

        :param overwrite_node_attr: bool, optional, default True
            Whether to overwrite attribute with the given group setting in
            self.lipid_attributes

        :param node_size_func: Callable or int, optional
            Used to compute node sizes if a function.
            Else set as constant node size.

        :param node_size_scale: tuple, optional, default (3, 12)
            Tuple of integers specifying the minimum and maximum
            of plotted node sizes. The actual values will be scaled
            to fit this range.

        :param node_cmap: str, optional, default 'tab10'
            Colourmap to use for edge colouring. Default for continuous
            attributes is 'viridis'

        :param edge_group_subset: tuple or list of size 2, optional
            Group-comparison to show in edge colouring/sizing

        :param node_group_subset: tuple or list of size 2, optional
            Group-comparison to show in node colouring/sizing

        :param save: bool, optional, default False,
            Whether to (only) save the plot

        :param show: bool, optional, default False,
            Whether to open the saved plot in the standard browser

        :param do_return: bool, optional, default False
            Whether to return the pyvis.network.Network object

        :param notebook: bool, optional, default False
            Whether the function is called from within a jupyter
            notebook (required to show interactive plot inline)

        :param net: pyvis.network.Network, optional
            If given this instance will be used instead of generating
            a new one within the function

        :param show_filters: str or list, optional
            Specifies one or multiple filters to activate. See pyvis
            documentation for available filters.

        :param as_directed: bool, optional, default False
            Whether to show a direted graph

        :param vmin: float, optional
            Minimum value of the legend (for continuous scales).
            If not given it will be set to 0 (for only positive
            arrays) or to -max(abs(min), abs(max)) (for zero-centred
            colours).
            Setting vmin can lead to un-centred colour scales!

        :param vmax: float, optional
            Maximum value of the legend (for continuous scales).
            If vmin is not specified, the minimum value will be
            either 0 (if all values are positive) or -vmax

        :param kwargs:
            keyword arguments to pass to pyvis.network.Network. Only
            releveant if net=None (default)

        Returns
        -------
        :return pyvis.network.Network or None
        """
        net = self.network_to_pyvis(
            edge_colour_attr=edge_colour_attr,
            overwrite_edge_attr=overwrite_edge_attr,
            edge_cmap=edge_cmap,
            edge_size_func=edge_size_func,
            edge_size_scale=edge_size_scale,
            node_cmap=node_cmap,
            node_size_func=node_size_func,
            node_colour_attr=node_colour_attr,
            overwrite_node_attr=overwrite_node_attr,
            node_size_scale=node_size_scale,
            edge_group_subset=edge_group_subset,
            node_group_subset=node_group_subset,
            net=net,
            as_directed=as_directed,
            vmin=vmin, vmax=vmax,
            **kwargs
        )
        if as_directed:
            net.set_edge_smooth("straight_cross")
        if show_filters is not None:
            if isinstance(show_filters, str):
                if show_filters == "all":
                    net.show_buttons(filter_=True)
                else:
                    net.show_buttons(filter_=[show_filters])
            else:
                net.show_buttons(filter_=show_filters)
        if do_return:
            return net
        if show:
            net.show(file)
            return
        if save:
            net.write_html(file, notebook)

    def _get_attr_abs_max_(self, attr: str, nodes: bool) -> float:
        attributes = self.lipid_attributes if nodes else self.interaction_attributes
        if attr not in attributes.keys():
            raise KeyError(
                f'{attr} not found in computed attribute list!'
            )
        if not isinstance(attributes[attr], dict):
            raise ValueError(
                f'{attr} have not been computed in a group-specific manner'
            )
        max_ = 0.
        for group, values in attributes[attr].items():
            if isinstance(values, SymmetricMatrix):
                cmax = np.nanmax(abs(values.data))
            elif isinstance(values, dict):
                cmax = np.nanmax(abs(np.array(list(values.values()))))
            else:
                raise ValueError(f'Wrong data type {type(values).__name__}')
            if cmax > max_:
                max_ = cmax
        return max_

    def add_network_colours(
        self, attributes: List[str],
        nodes: bool = True,
        cmaps: Dict[str, str] = None
    ) -> Dict[str, Dict[str, Any]]:
        max_ = None
        if nodes:
            generator = self._generate_node_colours
            setter = nx.set_node_attributes
            keys = self.network.nodes
        else:
            generator = self._generate_edge_colours_
            setter = nx.set_edge_attributes
            keys = self.network.edges
        if cmaps is None:
            cmaps = {"lipid_class": "tab20"}
        to_string = {}
        legends_maps = {}
        for attr in attributes:
            if attr in {'enzyme_gene_name', 'enzyme_uniprot', 'enzyme_id'}:
                continue
            if nodes:
                self._check_attributes_(None, attr)
            else:
                self._check_attributes_(attr, None)
            if self._attr_group_size_[attr] == 0 or \
                not hasattr(self, "unique_groups"):
                self._add_network_attribute_(
                    attr, nodes=nodes, group_subset=None,
                    overwrite=attr not in STATIC_NODE_PROPERTIES and attr not in STATIC_EDGE_PROPERTIES
                )
                if attr in to_string:
                    pass
                attr_colours = generator(
                    attr=attr, discrete=self._discrete_map_[attr],
                    cmap=cmaps.get(attr, "tab10"),
                    colours_to_hex=True
                )
                setter(
                    self.network,
                    dict(zip(keys, attr_colours[1])),
                    name=f"{attr}_colour"
                )
                legends_maps[f"{attr}_colour"] = attr_colours[0]
            elif self._attr_group_size_[attr] == 1:
                if not self._discrete_map_[attr]:
                    max_ = self._get_attr_abs_max_(attr, nodes)
                for group in self.unique_groups:
                    self._add_network_attribute_(
                        attr, nodes=nodes, group_subset=[group],
                        overwrite=attr not in STATIC_NODE_PROPERTIES and attr not in STATIC_EDGE_PROPERTIES,
                        add_group=True
                    )
                    if not self._discrete_map_[attr]:
                        for attr_name, vmax in zip(
                            [f"{attr}_colour_{group}_individual", f"{attr}_colour_{group}_common"],
                            [None, max_]):
                            attr_colours = generator(
                                attr,
                                self._discrete_map_[attr],
                                cmaps.get(attr, "tab10"),
                                group=group,
                                colours_to_hex=True,
                                group_attr=f"{attr}_{group}",
                                vmax=vmax
                            )
                            setter(
                                self.network,
                                dict(zip(keys, attr_colours[1])),
                                name=attr_name
                            )
                            legends_maps[attr_name] = attr_colours[0]
                    else:
                        attr_colours = generator(
                            attr,
                            self._discrete_map_[attr],
                            cmaps.get(attr, "tab10"),
                            group=group,
                            colours_to_hex=True,
                            group_attr=f"{attr}_{group}"
                        )
                        attr_name = f"{attr}_colour_{group}"
                        setter(
                            self.network,
                            dict(zip(keys, attr_colours[1])),
                            name=attr_name
                        )
                        legends_maps[attr_name] = attr_colours[0]
            else:
                if not self._discrete_map_[attr]:
                    max_ = self._get_attr_abs_max_(attr, nodes)
                for comb in self.comparisons:
                    if not self._discrete_map_[attr]:
                        for attr_name, vmax in zip(
                            [f"{attr}_colour_{_tuple_to_string_(comb)}_individual",
                             f"{attr}_colour_{_tuple_to_string_(comb)}_common"],
                            [None, max_]):
                            attr_colours = generator(
                                attr,
                                self._discrete_map_[attr],
                                cmaps.get(attr, "tab10"),
                                group=comb,
                                colours_to_hex=True,
                                group_attr=f"{attr}_{_tuple_to_string_(comb)}",
                                vmax=vmax
                            )
                            setter(
                                self.network,
                                dict(zip(keys, attr_colours[1])),
                                name=attr_name
                            )
                            legends_maps[attr_name] = attr_colours[0]
                    else:
                        self._add_network_attribute_(
                            attr, nodes=nodes, group_subset=comb,
                            overwrite=True, add_group=True,
                        )
                        attr_colours = generator(
                            attr,
                            self._discrete_map_[attr],
                            cmaps.get(attr, "tab10"),
                            group=comb,
                            colours_to_hex=True,
                            group_attr=f"{attr}_{_tuple_to_string_(comb)}"
                        )
                        attr_name = f"{attr}_colour_{_tuple_to_string_(comb)}"
                        setter(
                            self.network,
                            dict(zip(keys, attr_colours[1])),
                            name=attr_name
                        )
                        legends_maps[attr_name] = attr_colours[0]
        return legends_maps

    def add_node_sizes(
        self, attributes: List[str],
        scale: Tuple[float, float]
    ) -> Dict[str, Dict[float, float]]:
        max_ = None
        legends_maps = {}
        abs_ = False
        for attr in attributes:
            if attr in {'enzyme_gene_name', 'enzyme_uniprot', 'enzyme_id'}:
                continue
            if attr == "fold_changes":
                abs_ = False
            if attr in STATIC_NODE_PROPERTIES:
                node_names = self.network.nodes
            else:
                node_names = [self.reverse_mapping.get(node, node)
                              for node in self.network.nodes]
            self._check_attributes_(None, attr)
            if self._attr_group_size_[attr] == 0 or \
                    not hasattr(self, "unique_groups"):
                sizes, na_keys = _get_sizes_report_nas_(self.lipid_attributes[attr],
                                                        node_names)
                attr_size = self._scale_dict_(sizes, scale,
                                              map_as_ex=True)
                if na_keys:
                    for node, val in attr_size[0].items():
                        if np.isnan(val):
                            attr_size[0][node] = scale[0] / 2
                nx.set_node_attributes(
                    self.network, attr_size[0],
                    name=f"{attr}_size"
                )
                legends_maps[f"{attr}_size"] = attr_size[1]
            elif self._attr_group_size_[attr] == 1:
                if not self._discrete_map_[attr]:
                    max_ = self._get_attr_abs_max_(attr, nodes=True)
                for group in self.unique_groups:
                    sizes, na_keys = _get_sizes_report_nas_(self.lipid_attributes[attr][group],
                                                            node_names)
                    if not self._discrete_map_[attr]:
                        for attr_name, vmax in zip(
                                [f"{attr}_size_{group}_individual", f"{attr}_size_{group}_common"],
                                [None, max_]):
                            attr_size = self._scale_dict_(sizes, scale,
                                                          map_as_ex=True,
                                                          vmax=vmax)
                            if na_keys:
                                for node, val in attr_size[0].items():
                                    if np.isnan(val):
                                        attr_size[0][node] = scale[0] / 2
                            nx.set_node_attributes(
                                self.network, attr_size[0],
                                name=attr_name
                            )
                            legends_maps[attr_name] = attr_size[1]
                    else:
                        attr_size = self._scale_dict_(sizes, scale,
                                                      map_as_ex=True)
                        if na_keys:
                            for node, val in attr_size[0].items():
                                if np.isnan(val):
                                    attr_size[0][node] = scale[0] / 2
                        nx.set_node_attributes(
                            self.network, attr_size[0],
                            name=f"{attr}_size_{group}"
                        )
                        legends_maps[f"{attr}_size_{group}"] = attr_size[1]
            else:
                if not self._discrete_map_[attr]:
                    max_ = self._get_attr_abs_max_(attr, nodes=True)
                for comb in self.comparisons:
                    if self.lipid_attributes[attr].get(comb) is None:
                        comb = (comb[1], comb[0])
                    if attr == "fold_changes":
                        sizes, na_keys = _get_sizes_report_nas_(self.lipid_attributes[attr][comb],
                                                                node_names,
                                                                as_abs=True,
                                                                default=np.nan)
                    else:
                        sizes, na_keys = _get_sizes_report_nas_(self.lipid_attributes[attr][comb],
                                                                node_names,
                                                                default=np.nan)
                    if not self._discrete_map_[attr]:
                        for attr_name, vmax in zip(
                                [f"{attr}_size_{_tuple_to_string_(comb)}_individual",
                                 f"{attr}_size_{_tuple_to_string_(comb)}_common"],
                                [None, max_]):
                            if attr == 'fold_changes' or attr == "nlog_pvalues":
                                attr_size = self._scale_dict_(sizes, scale,
                                                              map_as_ex=True,
                                                              vmax=vmax,
                                                              map_data_names=self.lipid_mapping)
                            else:
                                attr_size = self._scale_dict_(sizes, scale,
                                                              map_as_ex=True,
                                                              vmax=vmax)
                            if na_keys:
                                for node, val in attr_size[0].items():
                                    if np.isnan(val):
                                        attr_size[0][node] = scale[0] / 2
                            nx.set_node_attributes(
                                self.network, attr_size[0],
                                name=attr_name
                            )
                            legends_maps[attr_name] = attr_size[1]
                    else:
                        attr_size = self._scale_dict_(sizes, scale,
                                                      map_as_ex=True)
                        if na_keys:
                            for node, val in attr_size[0].items():
                                if np.isnan(val):
                                    attr_size[0][node] = scale[0] / 2
                        nx.set_node_attributes(
                            self.network, attr_size[0],
                            name=f"{attr}_size_{_tuple_to_string_(comb)}"
                        )
                        legends_maps[f"{attr}_size_{_tuple_to_string_(comb)}"] = attr_size[1]
        return legends_maps

    def dynamic_network(
        self, node_colour_attributes: List[str],
        edge_colour_attributes: List[str],
        node_size_attributes: List[str],
        node_cmaps: Dict[str, str] = None,
        edge_cmaps: Dict[str, str] = None,
        node_size_scale: Tuple[int, int] = (10, 40),
        **kwargs
    ) -> DynamicVisParser:
        """
        Generating DynamicVisParser instance to plot the interactive network
        including all available visualisation options

        :param node_colour_attributes: List[str]
            List of node colour attributes to include. For available attributes
            use linex2.node_colour_options()
        :param edge_colour_attributes: List[str]
            List of edge colour attributes to include. For available attributes
            use linex2.edge_colour_options()
        :param node_size_attributes: List[str]
            List of node size attributes to include. For available attributes
            use linex2.node_size_options()
        :param node_cmaps: Dict[str, str]
            Specifying the color maps to use for individual node attributes. Dict
            keys are attributes, values are matplotlib-compatible color maps names.
        :param edge_cmaps: Dict[str, str]
            Specifying the color maps to use for individual edge attributes. Dict
            keys are attributes, values are matplotlib-compatible color maps names.
        :param node_size_scale: Tuple[int, int], default: (10, 40)
            Scale, as a 2-tuple of (min, max), indicating the minimum and maximum
            node size value for visualisation.
        :param kwargs
            Keyword arguments passed to the DynamicVisParser initialisation function
        """
        # adding static properties
        for node_prop in STATIC_NODE_PROPERTIES.difference('enzyme_gene_name', 'enzyme_uniprot', 'enzyme_id'):
            if node_prop not in node_colour_attributes:
                node_colour_attributes.append(node_prop)
            if node_prop != "lipid_class":
                if node_prop not in node_size_attributes:
                    node_size_attributes.append(node_prop)
        for edge_prop in STATIC_EDGE_PROPERTIES.difference('enzyme_gene_name', 'enzyme_uniprot', 'enzyme_id'):
            # reaction enzymes should not get any colours!
            if edge_prop != "enzyme_id":
                if edge_prop not in edge_colour_attributes:
                    edge_colour_attributes.append(edge_prop)
        # generating colours and adding to network attributes
        node_colour_legends = self.add_network_colours(node_colour_attributes,
                                                       nodes=True, cmaps=node_cmaps)
        edge_colour_legends = self.add_network_colours(edge_colour_attributes,
                                                       nodes=False, cmaps=edge_cmaps)
        node_size_legends = self.add_node_sizes(node_size_attributes, node_size_scale)

        dvp = DynamicVisParser(**kwargs)
        dvp.from_nx(self.network)
        if dvp.directed:
            dvp.set_edge_smooth("dynamic")
        dvp.generate_legend(
            node_colours=node_colour_legends,
            edge_colours=edge_colour_legends,
            node_sizes=node_size_legends,
            colours_to_hex=True
        )
        return dvp



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / 'README.md').read_text(encoding='utf-8')

setup(
    name='linex2',
    version='1.1.13',

    description='Lipid Network Explorer 2 (LINEX2) package. Python backend to the LINEX web-app',
    long_description=long_description,
    long_description_content_type='text/markdown',

    url='https://exbio.wzw.tum.de/linex/',
    project_urls={  # Optional
        'Source': 'https://gitlab.lrz.de/lipitum-projects/linex2_package',
        'Publication': "https://doi.org/10.1093/bib/bbac572"
    },

    author="Tim Daniel Rose, Nikolai Koehler, Olga Lazareva, Josch Konstantin Pauling",
    author_email="tim.rose@tum.de, nikolai.koehler@tum.de",

    license='AGPLv3',

    packages=find_packages(),
    install_requires=['networkx', "pandas", "matplotlib", "numpy", "pyvis",
                      "scikit-learn", "statsmodels", "jinja2", "seaborn"],
    python_requires=">=3.8",

    package_data={'linex2': ["templates/vis_main.html",
                                 "templates/vis_dynamic.html",
                                 "templates/lipid_colours.json",
                                 "data/*.csv", "data/*.txt"]},
    zip_safe=False,


    classifiers=[
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3"

    ]
)



# This file includes code from:
# LINEX webtool
# Copyright (C) 2021  LipiTUM group, Chair of experimental Bioinformatics, Technical University of Munich
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import pandas as pd
import numpy as np

# From package
from .utils import (
    correlations,
    partial_correlations,
    _matrix_pval_correction_,
    SymmetricMatrix,
    correlation_change,
    fold_changes,
    binary_test,
)
from .exceptions import SignificanceTestError
from .misc import _pandas_log_


def _correlation_computation_(self, attr: str,
                              method: str = "pearson",
                              estimator: str = "LedoitWolf",
                              significance: float = 0.05,
                              overwrite: bool = False,
                              correct_args: dict = None,
                              **kwargs):
    self._check_data_(check_groups=False)
    if self.groups is None:
        if not overwrite:
            if self.interaction_attributes.get(attr) is not None and self.interaction_attributes.get(attr):
                return
        if attr == "correlations":
            cors, pvals = correlations(self.data,
                                       method=method,
                                       **kwargs)
        else:
            cors, pvals = partial_correlations(self.data, estimator,
                                               **kwargs)
        # multiple test correction
        if not (pvals < 0).all():
            if correct_args is None:
                corr_pvals = _matrix_pval_correction_(pvals)
            else:
                corr_pvals = _matrix_pval_correction_(pvals, **correct_args)
            corr_pvals = pd.DataFrame(corr_pvals, cors.index, cors.columns)
            sig_mask = corr_pvals >= significance
            cors.values[sig_mask] = 0
        else:
            corr_pvals = pvals
        self.interaction_attributes[attr] = SymmetricMatrix(cors)
        self.interaction_attributes[f"{attr}_pvals"] = SymmetricMatrix(pd.DataFrame(corr_pvals,
                                                                                    cors.index,
                                                                                    cors.columns))
    else:
        if (self.interaction_attributes.get(attr) is None or
                not self.interaction_attributes.get(attr) or
                overwrite or
                isinstance(self.interaction_attributes.get(attr), SymmetricMatrix)
        ):
            self.interaction_attributes[attr] = {}
            self.interaction_attributes[f"{attr}_pvals"] = {}
            group_correlations = self.unique_groups.size * [False]
        else:
            group_correlations = [self.interaction_attributes[attr].get(group)
                                  is not None for group in self.unique_groups]
            if all(group_correlations):
                return
            elif (self.interaction_attributes.get(f"{attr}_pvals") is None or
                  overwrite or
                  isinstance(self.interaction_attributes.get(f"{attr}_pvals"), SymmetricMatrix)
            ):
                self.interaction_attributes[f"{attr}_pvals"] = {}

        for group in self.unique_groups[np.invert(group_correlations)]:
            group_data = self.data.loc[:, self.groups == group]
            if attr == "correlations":
                cors, pvals = correlations(group_data,
                                           method=method,
                                           **kwargs)
            else:
                cors, pvals = partial_correlations(group_data, estimator,
                                                   **kwargs)
            # multiple test correction
            if not (pvals < 0).all():
                if correct_args is None:
                    corr_pvals = _matrix_pval_correction_(pvals)
                else:
                    corr_pvals = _matrix_pval_correction_(pvals, **correct_args)
                sig_mask = corr_pvals >= significance
                cors.values[sig_mask] = 0
                cors = cors.fillna(0)
            else:
                corr_pvals = pvals
            self.interaction_attributes[attr][group] = SymmetricMatrix(cors)
            self.interaction_attributes[f"{attr}_pvals"][group] = SymmetricMatrix(pd.DataFrame(corr_pvals,
                                                                                               cors.index,
                                                                                               cors.columns))


def compute_correlations(self,
                         method: str = "pearson",
                         significance: float = 0.05,
                         overwrite: bool = False,
                         correct_args: dict = None,
                         **kwargs):
    """
    Computing correlations between all node pairs

    Parameters
    ----------
    :param method: str, optional, default 'pearson'
        Method to use for correlation calculation. Available
        options are: pearson, kendall and spearman

    :param significance: float, optional, default 0.05
        Significance threshold for corrected pvalues

    :param overwrite: bool, optional, default False
        If True previously computed correlations are overwritten

    :param correct_args: dict, optional
        Arguments to pass to statsmodels.stats.multitest.multipletests
        for multiple test correction

    :param kwargs:
        keyword arguments to pass to the respective scipy.stats method

    """
    self._correlation_computation_(attr="correlations",
                                   method=method, significance=significance,
                                   overwrite=overwrite, correct_args=correct_args,
                                   **kwargs)


def compute_partial_correlations(self,
                                 estimator: str = "GraphLasso",
                                 significance: float = 0.05,
                                 overwrite: bool = False,
                                 correct_args: dict = None,
                                 **kwargs):
    """
    Computing correlations between all node pairs

    Parameters
    ----------
    :param estimator: str, optional, default 'GraphLasso'
        Method to use for partial correlation estimation. Available
        options are: GraphLasso, LedoitWolf and empirical

    :param significance: float, optional, default 0.05
        Significance threshold for corrected p-values. Only relevant
        if sample to feature ratio is large enough to compute
        Fisher's z-transform

    :param overwrite: bool, optional, default False
        If True previously computed correlations are overwritten

    :param correct_args: dict, optional
        Arguments to pass to statsmodels.stats.multitest.multipletests
        for multiple test correction

    :param kwargs:
        keyword arguments to pass to the respective scipy.stats method


    Raises
    ------
    ValueError
        If groups or data are not of the correct type


    Warns
    -----
    If 'data' or 'groups' are already set but provided again
    """
    self._correlation_computation_(attr="partial_correlations",
                                   estimator=estimator, significance=significance,
                                   overwrite=overwrite, correct_args=correct_args,
                                   **kwargs)


def compute_correlation_changes(self, partial_corrs: bool = False,
                                data: pd.DataFrame = None,
                                groups: pd.Series = None,
                                pval_change: bool = True,
                                overwrite: bool = True):
    """
    Computing changes of partial correlations between different conditions

    Parameters
    ----------
    :param partial_corrs: bool, optional, default False
        Whether to use partial correlations

    :param data: pd.DataFrame or str, optional
        Used to overwrite self.data if given. Required
        if no lipid data was added before. If str it
        assumed to be a csv file with samples in columns.

    :param groups: pd.Series or str, optional
        Used to overwrite self.groups if given. Required
        if no groups were added before. If str it
        assumed to be a csv or txt file.

    :param pval_change: bool, optional, default True
        Whether to compute categories or absolute values

    :param overwrite: bool, optional, default False
        If True previously computed correlations are overwritten
    """
    self._check_data_()
    if partial_corrs:
        attr = "partial_correlations"
        attr_change = "partial_correlation_changes"
    else:
        attr = "correlations"
        attr_change = "correlation_changes"
    if self.interaction_attributes[attr] is None:
        raise ValueError(
            f"{attr} have not been computed yet!"
        )
    if self.interaction_attributes.get(attr_change) is None:
        self.interaction_attributes[attr_change] = {}

    self._set_data_groups_(data=data, groups=groups)
    # compute how correlations/partial correlations change between groups
    # => should be mostly referring to from significant to unsignificant
    # or vice-versa
    # TODO: which metrics to use
    for comb in self.comparisons:
        if not overwrite:
            if self.interaction_attributes[attr_change].get(tuple(comb)) is not None:
                continue
            elif self.interaction_attributes[attr_change].get(tuple(comb[0], comb[1])) is not None:
                continue
        if pval_change:
            if (self.interaction_attributes[f"{attr}_pvals"][comb[0]] < 0).all() or \
                    (self.interaction_attributes[f"{attr}_pvals"][comb[1]] < 0).all():
                raise SignificanceTestError(
                    "Too few samples for the given number of features to compute a fisher's z-transform."
                )
            corr_changes = correlation_change(
                self.interaction_attributes[attr][comb[0]],
                self.interaction_attributes[attr][comb[1]]
            )
        else:
            corr_changes = self.interaction_attributes[attr][comb[0]] - \
                           self.interaction_attributes[attr][comb[1]]
        self.interaction_attributes[attr_change][tuple(comb)] = corr_changes


def compute_fold_changes(self, data_is_log: bool = True, log_func=np.log2):
    """
    Computing (log) fold-changes

    Parameters
    ----------
    :param data_is_log: bool, optional, default False
        Indicating whether data is log-transformed

    :param log_func: callable, optional, default np.log2
        Which log function to use, in case data_is_log is False
    """
    # TODO: documentation
    self._check_data_()
    if self.groups is None or self.data is None:
        raise ValueError("fold changes cannot be computed when no groups or data is given")
    if self.lipid_attributes.get("fold_changes") is None:
        self.lipid_attributes["fold_changes"] = {
            tuple(comparison): fold_changes(self.data, self.groups,
                                            comparison, data_is_log,
                                            log_func).fillna(0).to_dict()
            for comparison in self.comparisons
        }
    else:
        for comparison in self.comparisons:
            comp_st = tuple(comparison)
            if self.lipid_attributes["fold_changes"].get(comp_st) is None:
                fcs = fold_changes(self.data, self.groups,
                                   comparison, data_is_log,
                                   log_func).fillna(0).to_dict()
                self.lipid_attributes["fold_changes"][comp_st] = fcs


def compute_pvalues(
        self, round_vals: bool = False,
        method: str = "ttest"
):
    """
    Computing p-values

    Parameters
    ----------
    :param round_vals: bool, optional, default False
        If True pvalues will be rounded to four digits

    :param method: str, optional, default 'ttest'
        Which test to use. Available options are:
        'ttest', 'wilcoxon', '' and ''
    """
    self._check_data_()
    # computing p-values
    if self.lipid_attributes.get("pvalues") is None:
        pvals = {tuple(comparison): binary_test(self.data, self.groups,
                                                comparison, method=method)
                 for comparison in self.comparisons}
        nlog_pvals = {(comp[0], comp[1]): (-_pandas_log_(parr)).to_dict()
                      for comp, parr in pvals.items()}

        self.lipid_attributes["pvalues"] = {key: val.to_dict() for key, val in pvals.items()}
        self.lipid_attributes["nlog_pvalues"] = nlog_pvals
    else:
        for comparison in self.comparisons:
            comp_st = tuple(comparison)
            if self.lipid_attributes["pvalues"].get(comp_st) is None:
                pval = binary_test(self.data, self.groups,
                                   comparison)
                self.lipid_attributes["pvalues"][comp_st] = pval.to_dict()
                self.lipid_attributes["nlog_pvalues"][comp_st] = (-_pandas_log_(pval)).to_dict()



# ###
# Classes for lipids that will be used for the dynamic network creation
# ###
from __future__ import annotations
import warnings
from typing import List, Tuple
import itertools
from .exceptions import MolecularSpeciesError


def flatten(l):
    return [item for sublist in l for item in sublist]


class DBPos:

    __slots__ = ["loc", "geom"]

    def __init__(self, loc: int, geom: str):
        self.loc = loc
        self.geom = geom.upper()

    def __str__(self):
        return str(self.loc) + self.geom

    def __lt__(self, other):
        if self.loc < other.loc:
            return True
        elif self.loc == other.loc:
            return self.geom < other.geom
        else:
            return False

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return self.loc == other.loc


class FA:

    __slots__ = ["__C", "__DB", "__OH", "__modifications", "__is_long_chain_base", "__is_ether_bond",
                 "__ether_type", "__db_positions"]

    def __init__(self, c: int, db: int, oh: int = None, modifications: list = None,
                 is_long_chain_base: bool = False, is_ether_bond: bool = False, ether_type: str = None,
                 db_positions: List[DBPos] = None):
        """
        Fatty acid class. Describes a fatty acid that can be part of a Lipid
        :param c: int, Number of carbons
        :param db: int, Number of double bonds
        :param oh: int, Number of hydroxylations, can be None.
        :param modifications: List of string, Other modifications of the fatty acid
        """
        self.__C = c
        self.__DB = db
        if oh is None:
            self.__OH = 0
        else:
            self.__OH = oh
        self.__modifications = modifications
        self.__is_long_chain_base = is_long_chain_base
        self.__is_ether_bond = is_ether_bond
        if self.__is_ether_bond and ether_type is None:
            self.__ether_type = "O-"
        else:
            self.__ether_type = ether_type

        if self.__ether_type is not None:
            self.__ether_type = self.__ether_type.upper()

        if modifications is not None:
            self.__modifications.sort()

        if self.__is_long_chain_base and self.__is_ether_bond:
            warnings.warn("Fatty acid is long chain base and ether bond at once.")

        if db_positions is None:
            self.__db_positions = None
        else:
            if len(db_positions) != db:
                warnings.warn("DB positions does not match the number of double bonds")
            self.__db_positions = sorted(db_positions)

    def length(self) -> int:
        return self.__C

    def db_positions(self) -> List[DBPos]:
        return self.__db_positions

    def db_index(self) -> int:
        return self.__DB

    def hydroxylations(self) -> int:
        return self.__OH

    def modifications(self) -> List:
        return self.__modifications

    def is_long_chain_base(self) -> bool:
        return self.__is_long_chain_base

    def is_ether_bond(self) -> bool:
        return self.__is_ether_bond

    def get_ether_type(self) -> str:
        return self.__ether_type

    def _db_pos_equal(self, other: FA) -> bool:
        if self.db_positions() is None or other.db_positions() is None:
            return True
        else:
            return self.__db_positions == other.db_positions()

    def __eq__(self, other) -> bool:
        if not isinstance(other, FA):
            raise TypeError("Only comparisons between FAs allowed.")
        return (self.__C == other.length()) & (self.__DB == other.db_index()) &\
               (self.__OH == other.hydroxylations()) & (self.__modifications == other.modifications()) &\
               (self.__is_long_chain_base == other.is_long_chain_base()) &\
               (self.__is_ether_bond == other.is_ether_bond()) &\
            self._db_pos_equal(other)

    def __str__(self) -> str:
        out_str = ""
        if self.__is_ether_bond:
            out_str += self.__ether_type
        # if self.__is_long_chain_base:
        #     out_str += "d"
        if self.modifications() is not None:
            if len(self.modifications()) > 0:
                for mod in self.modifications():
                    out_str += str(mod + '-')
        out_str += f"{self.__C}:{self.__DB}"
        if self.__db_positions is not None:
            out_str += "("
            out_str += ",".join([str(x) for x in self.db_positions()])
            out_str += ")"
        if self.__OH:
            if self.__OH > 1:
                out_str += f";O{self.__OH}"
            else:
                out_str += f";O"
        return out_str

    def __hash__(self):
        return hash("l" + str(self)) # Mark as LCB for hashing

    def __repr__(self) -> str:
        return self.__str__()

    def __lt__(self, other) -> bool:
        if (not self.__is_long_chain_base) and other.is_long_chain_base():
            return False
        elif self.__is_long_chain_base and (not other.is_long_chain_base()):
            return True
        elif (not self.__is_ether_bond) and other.is_ether_bond():
            return False
        elif self.__is_ether_bond and (not other.is_ether_bond()):
            return True
        else:
            if self.__C < other.length():
                return True
            elif self.__C > other.length():
                return False
            else:
                if self.__DB < other.db_index():
                    return True
                elif self.__DB > other.db_index():
                    return False
                else:
                    myoh = 0 if self.__OH is None else self.__OH
                    otheroh = 0 if other.hydroxylations() is None else other.hydroxylations()
                    if myoh < otheroh:
                        return True
                    elif myoh > otheroh:
                        return False

    def __le__(self, other) -> bool:
        if (not self.__is_long_chain_base) and other.is_long_chain_base():
            return False
        elif self.__is_long_chain_base and (not other.is_long_chain_base()):
            return True
        if (not self.__is_ether_bond) and other.is_ether_bond():
            return False
        elif self.__is_ether_bond and (not other.is_ether_bond()):
            return True
        else:
            if self.__C <= other.length():
                return True
            elif self.__C > other.length():
                return False
            else:
                if self.__DB <= other.db_index():
                    return True
                elif self.__DB > other.db_index():
                    return False
                else:
                    myoh = 0 if self.__OH is None else self.__OH
                    otheroh = 0 if other.hydroxylations() is None else other.hydroxylations()
                    if myoh <= otheroh:
                        return True
                    elif myoh > otheroh:
                        return False

    def modified_fa(self, c, db, oh, added_db_positions: List[DBPos] = None) -> FA:
        """
        Adding increasing length and unsaturation -> The other way will introduce problems with DB positions

        """
        if db == 0:
            return FA(self.__C + c, self.__DB + db, self.__OH + oh, self.__modifications, self.__is_long_chain_base,
                      self.__is_ether_bond, self.__ether_type, db_positions=self.__db_positions)
        elif db > 0 and added_db_positions is None:
            return FA(self.__C + c, self.__DB + db, self.__OH + oh, self.__modifications, self.__is_long_chain_base,
                      self.__is_ether_bond, self.__ether_type, db_positions=None)
        else:
            new_db_positions = sorted(list(set(self.__db_positions + added_db_positions)))
            if len(new_db_positions) < (len(self.__db_positions) + len(added_db_positions)):
                warnings.warn("Duplicated Double bond positions added")
            return FA(self.__C + c, self.__DB + db, self.__OH + oh, self.__modifications, self.__is_long_chain_base,
                      self.__is_ether_bond, self.__ether_type, db_positions=new_db_positions)

    def as_ether(self, ether_type: str) -> FA:
        return FA(self.__C, self.__DB, self.__OH, self.__modifications, self.__is_long_chain_base,
                  is_ether_bond=True, ether_type=ether_type, db_positions=self.__db_positions)

    def as_not_ether(self) -> FA:
        return FA(self.__C, self.__DB, self.__OH, self.__modifications, self.__is_long_chain_base,
                  is_ether_bond=False, db_positions=self.__db_positions)


class Lipid:

    __slots__ = ["__lipid_class", "__head_group", "__fa_spots", "__backbone", "__non_fa_modifications", "__ids",
                 "_is_molecular_species", "__sum_length", "__sum_dbs", "__sum_ohs", "__sum_fa_modifications",
                 "__fatty_acids", "__number_fas", "__has_long_chain_base", "__has_ether_bond",
                 "__ether_type", "__dataname", "__converted_to_mol_spec", "__category", "__div_factor"]

    def __init__(self, lipid_class: str,
                 head_group: str,
                 is_molecular_species: bool,
                 fa_spots: int,
                 backbone: str = "glycerol",
                 dataname: str = None,
                 non_fa_modifications: list = None,
                 ids: dict = None,
                 fatty_acids: List[FA] = None,
                 sum_length: int = None,
                 sum_dbs: int = None,
                 sum_ohs: int = None,
                 sum_fa_modifications: list = None,
                 number_fas: int = None,
                 has_long_chain_base: int = None,
                 has_ether_bond: int = None,
                 ether_type: str = None,
                 converted_to_mol_spec: bool = False,
                 category: str = "Glycerolipids"):
        """
        Lipid class. Stores all necessary attributes of a lipid. Can be used to apply reaction rules to it.
        :param lipid_class: String, class of a lipid
        :param head_group: String, head group of a lipid
        :param is_molecular_species: Bool
        :param fa_spots: Int Number of possible fatty acids connected to this lipid
        :param backbone: String, Default="glycerol" backbone of the lipid (Important for reaction rules)
        :param non_fa_modifications: List of strings, Modification of the lipid (not related to faty acids)
        :param ids: dict, Database IDs of the lipid, with database as key and value as ID
        :param fatty_acids: list of fatty acids (class FA), required if is_molecular_species==True
        :param sum_length: Int, Number of carbons of the lipids, required if is_molecular_species==False
        :param sum_dbs: Int, Number of double bonds, required if is_molecular_species==False
        :param sum_ohs: Int, Number of hydroxylations of fatty acids, required if is_molecular_species==False, can be None
        :param sum_fa_modifications: List of strings, Modifications of fatty acids, required if is_molecular_species==False
        :param number_fas: Int, number of fatty acids attached to this lipid, required if is_molecular_species==False
        :param has_long_chain_base: Int, Number of long chain bases in this lipid, required if is_molecular_species==False
        """
        self.__lipid_class = lipid_class
        self.__head_group = head_group
        self.__fa_spots = fa_spots
        self.__backbone = backbone
        self.__category = category
        self.__non_fa_modifications = non_fa_modifications
        self.__ids = ids
        self._is_molecular_species = is_molecular_species
        self.__sum_length = sum_length
        self.__sum_dbs = sum_dbs
        self.__sum_ohs = sum_ohs
        self.__sum_fa_modifications = sum_fa_modifications
        self.__fatty_acids = fatty_acids
        self.__number_fas = number_fas
        self.__has_long_chain_base = has_long_chain_base
        self.__has_ether_bond = has_ether_bond
        self.__dataname = dataname
        self.__converted_to_mol_spec = converted_to_mol_spec
        if self.__has_ether_bond and (ether_type is None or ether_type==""):
            self.__ether_type = "O-"
        else:
            self.__ether_type = ether_type

        if is_molecular_species:
            if fatty_acids is None:
                raise AttributeError("if is_molecular_species==True, "
                                     "a list of fatty acids (class FA) must be provided.")
            else:
                self.__sum_length = None
                self.__sum_dbs = None
                self.__sum_ohs = None
                self.__sum_fa_modifications = None
                self.__number_fas = len(fatty_acids)
                self.__has_long_chain_base = None
                self.__has_ether_bond = None
        else:
            if (sum_length is None) or (sum_dbs is None) or (sum_ohs is None) or \
                    (sum_fa_modifications is None) or (number_fas is None) or (has_long_chain_base is None) or \
                    (has_ether_bond is None):
                raise AttributeError("if is_molecular_species==False, a all other arguments must be provided.")
            else:
                self.__fatty_acids = None

    def get_dataname(self):
        return self.__dataname

    def set_dataname(self, name):
        self.__dataname = name

    def is_molecular_species(self) -> bool:
        return self._is_molecular_species

    def get_headgroup(self) -> str:
        return self.__head_group

    def get_lipid_class(self) -> str:
        return self.__lipid_class

    def sum_length(self) -> int:
        if self._is_molecular_species:
            return sum([x.length() for x in self.__fatty_acids])
        else:
            return self.__sum_length

    def sum_dbs(self) -> int:
        if self._is_molecular_species:
            return sum([x.db_index() for x in self.__fatty_acids])
        else:
            return self.__sum_dbs

    def sum_ohs(self) -> int:
        if self._is_molecular_species:
            return sum([x.hydroxylations() for x in self.__fatty_acids if x.hydroxylations() is not None])
        else:
            return self.__sum_ohs

    def sum_fa_modifications(self) -> List[str]:
        if self._is_molecular_species:
            return flatten([x.modifications() for x in self.__fatty_acids if x.modifications() is not None])
        else:
            return self.__sum_fa_modifications

    def get_non_fa_modifications(self) -> List:
        return self.__non_fa_modifications

    def get_number_fas(self) -> int:
        if self._is_molecular_species:
            return len(self.__fatty_acids)
        return self.__number_fas

    def get_fa_spots(self) -> int:
        return self.__fa_spots

    def get_fa_lengths(self) -> List[int]:
        if self._is_molecular_species:
            return [x.length() for x in self.__fatty_acids]
        else:
            print("Lipid is not a molecular species (get_fa_lengths)")
            return []

    def get_fa_dbs(self) -> List[int]:
        if self._is_molecular_species:
            return [x.db_index() for x in self.__fatty_acids]
        else:
            print("Lipid is not a molecular species (get_fa_dbs)")
            return []

    def get_fa_ohs(self) -> List[int]:
        if self._is_molecular_species:
            return [x.hydroxylations() for x in self.__fatty_acids]
        else:
            print("Lipid is not a molecular species (get_fa_ohs)")
            return []

    def get_modifications(self):
        if self._is_molecular_species:
            return [x.modifications() for x in self.__fatty_acids]
        else:
            print("Lipid is not a molecular species (get_modifications)")
            return self.__sum_fa_modifications

    def get_fas(self) -> List[FA]:
        if self._is_molecular_species:
            return self.__fatty_acids
        else:
            print("Lipid is not a molecular species (get_fas)")
            return []

    def get_backbone(self) -> str:
        return self.__backbone

    def get_category(self) -> str:
        return self.__category

    def get_ids(self) -> dict:
        return self.__ids

    def get_fa_long_chain_base(self) -> List[bool]:
        if self._is_molecular_species:
            if len(self.__fatty_acids) == 0:
                return []
            return [x.is_long_chain_base() for x in self.__fatty_acids]
        else:
            print("Lipid is not a molecular species")
            return []

    def get_number_long_chain_base(self) -> int:
        if self._is_molecular_species:
            if len(self.__fatty_acids) == 0:
                return 0
            return sum([x.is_long_chain_base() for x in self.__fatty_acids])
        else:
            return self.__has_long_chain_base

    def get_ether(self) -> int:
        if self._is_molecular_species:
            if len(self.__fatty_acids) == 0:
                return 0
            return sum([int(x.is_ether_bond()) for x in self.__fatty_acids])
        else:
            return self.__has_ether_bond

    def get_fa_ether(self) -> List[bool]:
        if self._is_molecular_species:
            if len(self.__fatty_acids) == 0:
                return []
            return [x.is_ether_bond() for x in self.__fatty_acids]
        else:
            print("Lipid is not a molecular species")
            return []

    def get_ether_type(self) -> str:
        if self._is_molecular_species:
            if len(self.__fatty_acids) == 0:
                return []
            tmp = [x.get_ether_type() for x in self.__fatty_acids if x.get_ether_type() not in [None, ""]]
            if len(tmp) == 0:
                return ""
            elif len(tmp) == 1:
                return tmp[0]
            else:
                return str(tmp)
        else:
            return self.__ether_type

    def get_molecular_species(self, fatty_acids: List[FA]) -> List[Lipid]:
        if not all([isinstance(x, FA) for x in fatty_acids]):
            raise TypeError("All elements in 'fatty_acids' must be of type FA.")
        if self._is_molecular_species:
            warnings.warn("Lipid is already a molecular species")

        combinations = list()
        if self.get_ether():
            new_fa = fatty_acids.copy() + [x.as_ether(self.get_ether_type()) for x in fatty_acids]
            combs = itertools.combinations_with_replacement(new_fa, self.get_number_fas())
        else:
            combs = itertools.combinations_with_replacement(fatty_acids, self.get_number_fas())

        for i in combs:
            tmp_l = Lipid(lipid_class=self.get_lipid_class(),
                          head_group=self.__head_group,
                          is_molecular_species=True,
                          fa_spots=self.__fa_spots,
                          dataname=self.__dataname,
                          non_fa_modifications=self.__non_fa_modifications,
                          ids=self.__ids,
                          backbone=self.__backbone,
                          fatty_acids=list(i),
                          converted_to_mol_spec=True,
                          category=self.__category)

            if molecular_equals_sum_species(tmp_l, self):
                combinations.append(tmp_l)

        if len(combinations) == 0:
            raise MolecularSpeciesError(f"No molecular species could be inferred for {self.sum_species_str()}.\n"
                                        f"Try to add more possible fatty acids | {self.sum_species_str()}")
        return combinations

    def __str__(self) -> str:
        if self.is_molecular_species():
            if " O-" in self.__lipid_class:
                clsstr = self.__lipid_class.replace(" O-", "")
                return f"{clsstr}({'_'.join([str(x) for x in sorted(self.__fatty_acids)])})"
            else:
                return f"{self.__lipid_class}({'_'.join([str(x) for x in sorted(self.__fatty_acids)])})"
        else:
            return self.sum_species_str()

    def sum_species_str(self) -> str:
        if self.get_ether() and " O-" in self.__lipid_class:
            clsstr = self.__lipid_class.replace(" O-", "")
            out_str = f"{clsstr}("
        else:
            out_str = f"{self.__lipid_class}("
        if self.sum_fa_modifications() is not None:
            if len(self.sum_fa_modifications()) > 0:
                for mod in self.sum_fa_modifications():
                    out_str += str(mod + '-')
        if self.get_ether():
            out_str += self.get_ether_type()
        # if self.get_number_long_chain_base():
        #     out_str += "d"
        out_str += f"{self.sum_length()}:{self.sum_dbs()}"
        if self.sum_ohs():
            if self.sum_ohs() > 1:
                out_str += f";O{self.sum_ohs()}"
            else:
                out_str += f";O"
        return out_str + ")"

    def __repr__(self) -> str:
        return self.__str__()

    def __hash__(self):
        return hash(self.__str__())

    def converted_to_mol_spec(self) -> bool:
        return self.__converted_to_mol_spec

    def set_converted_to_mol_spec(self, converted=True):
        self.__converted_to_mol_spec = converted

    def get_native_string(self) -> str:
        if self.converted_to_mol_spec():
            return self.sum_species_str()
        else:
            return self.__str__()

    def set_div_factor(self, fac: float):
        self.__div_factor = fac

    def get_div_factor(self):
        return self.__div_factor


def molecular_equals_sum_species(ms: Lipid, ss: Lipid) -> bool:
    if ms.is_molecular_species() and not ss.is_molecular_species():
        if ms.get_ether() or ss.get_ether():
            return (ms.get_lipid_class() == ss.get_lipid_class()) and \
                   (ms.sum_length() == ss.sum_length()) and \
                   (ms.sum_dbs() == ss.sum_dbs()) and \
                   (ms.sum_ohs() == ss.sum_ohs()) and \
                   (ms.get_headgroup() == ss.get_headgroup()) and \
                   (ms.get_fa_spots() == ss.get_fa_spots()) and \
                   (ms.get_number_fas() == ss.get_number_fas()) and \
                   (ms.get_ether() == ss.get_ether()) and \
                   (ms.get_number_long_chain_base() == ms.get_number_long_chain_base()) and \
                   (ms.get_ether_type() == ss.get_ether_type())
        else:
            return (ms.get_lipid_class() == ss.get_lipid_class()) and \
                   (ms.sum_length() == ss.sum_length()) and \
                   (ms.sum_dbs() == ss.sum_dbs()) and \
                   (ms.sum_ohs() == ss.sum_ohs()) and \
                   (ms.get_headgroup() == ss.get_headgroup()) and \
                   (ms.get_fa_spots() == ss.get_fa_spots()) and \
                   (ms.get_number_fas() == ss.get_number_fas()) and \
                   (ms.get_ether() == ss.get_ether()) and \
                   (ms.get_number_long_chain_base() == ms.get_number_long_chain_base())
    else:
        warnings.warn("Comparison requires a molecular and a sum species lipid.")
        return False


if __name__ == "__main__":
    fa1 = FA(20, 4, None)
    fa2 = FA(18, 0, 0)
    fa3 = FA(22, 4, 0, is_ether_bond=True, ether_type="P-")
    lcb1 = FA(18, 0, 1, is_long_chain_base=True)

    pe1 = Lipid("PE", "PE", is_molecular_species=True, fa_spots=2, fatty_acids=[fa1, fa2])
    pe2 = Lipid("PE-O", "PE", is_molecular_species=True, fa_spots=2, fatty_acids=[fa1, fa3])
    sm1 = Lipid("SM", "PC", is_molecular_species=True, fa_spots=2, fatty_acids=[fa1, lcb1])

    print("++ Example molecular species++")
    print(pe1)
    print(pe2)
    print(sm1)

    pc1 = Lipid("PC", "PC", is_molecular_species=False, fa_spots=2, sum_length=34, sum_dbs=1, sum_ohs=0,
                sum_fa_modifications=[], number_fas=2, has_long_chain_base=0, has_ether_bond=0)

    print("")
    print("++ Example non molecular species++")
    print(pc1)

    print("")
    print("FA comparisson")
    print("FA(20, 4, 0) == FA(20, 4, 0): ", FA(20, 4, 0) == FA(20, 4, 0))
    print("FA(18, 1, 0) == FA(20, 4, 0): ", FA(18, 1, 0) == FA(20, 4, 0))

    # print(pc1.get_combinations_of_molecular_species([fa1, fa2, lcb1]))



import numpy as np
import operator
from sklearn.cluster import KMeans
import networkx as nx
import seaborn as sns
import pandas as pd
import matplotlib.pyplot as plt


flatten = lambda l: [item for sublist in l for item in sublist]

split_str = "#"


def convergence_plot(scores):
    """
    Shows the convergence plot

    Attributes:
    -----------
    scores - the output of run_search() function

    output - directory name where results should be saved
    """

    plt.figure(figsize=(10, 6))

    sns.set(style="whitegrid")
    plt.rc('font', size=13)  # controls default text sizes
    plt.rc('axes', titlesize=13)  # fontsize of the axes title
    plt.rc('xtick', labelsize=13)  # fontsize of the tick labels
    plt.rc('ytick', labelsize=13)  # fontsize of the tick labels
    plt.rc('legend', fontsize=13)
    wg = pd.DataFrame(scores, columns=["score"])
    ax = sns.lineplot(data=wg, palette="tab10", linewidth=2.5)
    ax.set(xlabel="Iterations")
    ax.set(ylabel="Score")
    plt.show()


class LSOprimizer:
    def __init__(self,
                 G: nx.Graph,
                 L_min: int,
                 L_max: int,
                 T: float = 20,
                 max_iter: int = 100,
                 reaction_penalty: float = 1,
                 plot: bool = True,
                 opt_pat=None,
                 k=2,
                 init_size=None,
                 seed=None,
                 verbose=True,
                 node_scoring=False):
        """
        Given a graph G and gene expression matrix GE finds the optimal subnetwork in G of size at least L_min and
        at most L_max that can provide the optimal patients clustering in k clusters.
        :param G: networkX graph with PPI network
        :param L_min: minimal desired solution subnetwork size
        :param L_max: maximal desired solution subnetwork size
        :param T: temprature parameter for SA
        :param max_iter: maximal allowed number of iterations
        :param plot: convergence plot (True/False)
        :param opt_pat: patients labels (if provided, patients clustering won't be performed
        :param k: nmber of clusters
        :param init_size: initial subnetwork size (default L_max *2)
        :param seed: seed
        :param verbose: True/False
        """
        self.G = G
        self.T = T
        self.L_min = L_min
        self.L_max = L_max
        self.max_iter = max_iter
        self.reaction_penalty = reaction_penalty
        self.plot = plot
        if opt_pat is None:
            self.opt_pat = []
        else:
            self.opt_pat = opt_pat
        self.k = k
        if init_size is None:
            self.init_size = L_max
        else:
            self.init_size = init_size
        self.seed = seed
        self.verbose = verbose
        self.genes = list(self.G.nodes)
        self.nodescoring = node_scoring

    def APUtil(self, u, visited, ap, parent, low, disc, nodes, Time=0):
        """
        A recursive function that find articulation points
        using DFS traversal
        :param u: the vertex to be visited next
        :param visited: keeps tract of visited vertices
        :param ap: stores articulation points
        :param parent: stores parent vertices in DFS tree
        :param low: low value
        :param disc: stores discovery times of visited vertices
        :param nodes: current node set

        for more details: https://www.geeksforgeeks.org/articulation-points-or-cut-vertices-in-a-graph/
        """

        # Count of children in current node
        children = 0

        # Mark the current node as visited and print it
        visited[u] = True

        # Initialize discovery time and low value
        disc[u] = Time
        low[u] = Time
        Time += 1

        # for all the vertices adjacent to this vertex
        for v in self.G.neighbors(u):
            # If v is not visited yet, then make it a child of u
            # in DFS tree and recur for it
            if v in nodes:
                if not visited[v]:
                    parent[v] = u
                    children += 1
                    self.APUtil(v, visited, ap, parent, low, disc, nodes, Time)

                    # Check if the subtree rooted with v has a connection to
                    # one of the ancestors of u
                    low[u] = min(low[u], low[v])

                    # u is an articulation point in following cases
                    # (1) u is root of DFS tree and has two or more children.
                    if parent[u] == -1 and children > 1:
                        ap[u] = True

                    # (2) If u is not root and low value of one of its child is more
                    # than discovery value of u.
                    if parent[u] != -1 and low[v] >= disc[u]:
                        ap[u] = True

                        # Update low value of u for parent function calls
                elif v != parent[u]:
                    low[u] = min(low[u], disc[v])

                    # The function to do DFS traversal. It uses recursive APUtil()

    def is_AP(self, nodes):
        """
        Checks which nodes in the given set of nodes can NOT be removed without breaking
        disconnecting the induced subgraph
        :param nodes: set of nodes that make an induced subgraph of G
        :return: dictionary where each key is a node and each value indicates if a node is
        removable (articulation point)
        """
        visited = dict()
        disc = dict()
        low = dict()
        parent = dict()
        ap = dict()
        for node in nodes:
            visited[node] = False
            disc[node] = float("Inf")
            low[node] = float("Inf")
            parent[node] = -1
            ap[node] = False

        # Call the recursive helper function
        # to find articulation points
        # in DFS tree rooted with vertex 'i'
        for node in nodes:
            if not visited[node]:
                self.APUtil(node, visited, ap, parent, low, disc, set(nodes))

        return ap

    def score(self, nodes):
        """
        scores  given solution which is defined as a subnetwork and patient clusters
        :param nodes: list of nodes used in the solution
        :return: objective function value
        """

        sub = nx.subgraph(self.G, nodes)
        scores = []
        reacs = []
        if not self.nodescoring:
            for ed in sub.edges:
                tmp_edge_data = sub.get_edge_data(*ed)
                scores.append(tmp_edge_data["objective"])
                reacs.append(tmp_edge_data["enzyme_raw_id"])
            return np.sum(scores) / \
                (len(scores) +
                 (self.reaction_penalty * len(set(reacs)))
                 )
        else:
            scores = [val for key, val in dict(sub.nodes(data='objective')).items()]
            if self.reaction_penalty > 0:
                n_enz = len(set(flatten([val for key, val in dict(sub.nodes(data='enzymes')).items()])))
                enz_pen = 1 if n_enz < 1 else n_enz
                # print(enz_pen)
                return np.mean(scores) / (enz_pen * self.reaction_penalty)
            else:
                # print(set(flatten([val for key, val in dict(sub.nodes(data='enzymes')).items()])))
                #n_enz = len(set(flatten([val for key, val in dict(sub.nodes(data='enzymes')).items()])))
                #enz_pen = 1 if n_enz < 1 else n_enz
                #print(np.mean(scores) - (enz_pen * 1.))
                #print(np.mean(scores))
                #print()
                return np.mean(scores)

    @staticmethod
    def static_score(g, reaction_penalty, nodes):
        sub = nx.subgraph(g, nodes)
        scores = []
        reacs = []
        for ed in sub.edges:
            tmp_edge_data = sub.get_edge_data(*ed)
            scores.append(tmp_edge_data["objective"])
            reacs.append(tmp_edge_data["enzyme_raw_id"])

        return np.sum(scores) / \
            (len(scores) +
             (reaction_penalty * len(set(reacs)))
             )

    def dfs(self, node, d, visited=None):
        """
        Recursive DFS
        :param node: starting node
        :param d: length of required subnetwork
        :param visited: should be left empty
        :return: a list of connected nodes of length d
        """

        if visited is None:
            visited = []
        if node not in visited and len(visited) < d:
            visited.append(node)
            for neighbour in self.G.neighbors(node):
                self.dfs(neighbour, d, visited)
        if len(visited) == d:
            return visited

    def get_candidates(self, nodes):
        """
        Outputs first-degree neighbours of given nodes in graph G
        :param nodes: list of nodes that form a subnetwork/solution
        :return: list of first neighbour nodes
        """
        subst_candidates = flatten([[n for n in self.G.neighbors(x)] for x in nodes])
        subst_candidates = set(subst_candidates).difference(set(nodes))
        return subst_candidates

    def insertion(self, nodes):
        """
        Scores all possible insertions
        :param nodes: current solution
        :return: dictionary where key are possible insertions and values are scores
        """
        results = dict()
        size = len(nodes)
        if size < self.L_max:
            candidates = self.get_candidates(nodes)
            for c in candidates:
                nodes_new = nodes + [c]
                sc = self.score(nodes_new)
                results["i" + split_str + c] = sc
        return results

    def deletion(self, nodes, AP):
        """
        Scores all possible deletions
        :param nodes: current solution
        :param AP: articulation points (can't be removed since they separate the subnetwork)
        :return: dictionary where key are possible deletions and values are scores
        """
        results = dict()
        size = len(nodes)

        if size > self.L_min:
            for node in nodes:
                if not AP[node]:
                    nodes_new = list(set(nodes).difference({node}))
                    sc = self.score(nodes_new)
                    results["d" + split_str + str(node)] = sc
        return results

    def subst(self, nodes, AP):
        """
        Scores all possible substitutions
        :param nodes: current solution
        :param AP: articulation points (can't be removed since they separate the subnetwork)
        :return: dictionary where key are possible substitutions and values are scores
        """
        results = dict()
        size = len(nodes)
        if (size < self.L_max) and (size > self.L_min):
            for node in nodes:
                without_node = set(nodes) - {node}
                candidates = self.get_candidates(list(without_node))
                candidates = candidates - {node}
                for c in candidates:
                    if AP[node]:
                        nodes_new = list(without_node.union({c}))
                        if self.is_connected(nodes_new):
                            sc = self.score(nodes_new)
                            results["s" + split_str + str(node) + split_str + str(c)] = sc

                    else:
                        nodes_new = list(without_node.union({c}))
                        sc = self.score(nodes_new)
                        results["s" + split_str + str(node) + split_str + str(c)] = sc

        return results

    def is_connected(self, nodes):
        """
        Checks if a subgraph of G that consists of the given nodes is connected
        :param nodes: list of nodes
        :return: bool
        """
        # sg = self.G.new_vertex_property("bool")
        sg = self.G.subgraph(nodes)
        # g = gt.GraphView(self.G, vfilt=sg)

        return nx.is_connected(sg)

        # comp, _ = gt.label_components(g, vprop=sg)
        # if len(set(comp.a[nodes])) > 1:
        #     return False
        # else:
        #     return True

    @staticmethod
    def do_action_nodes(action, nodes):
        """
        Updates the set of nodes given the action
        :param action: a key from the results dictionary that has a description of an action
        :param nodes: previous solution
        :return: new set of nodes
        """
        if len(action.split(split_str)) == 2:  # inserion or deletion
            act, node = action.split(split_str)
            node = node
            if act == "i":
                nodes = nodes + [node]
            else:
                nodes = list(set(nodes).difference({node}))
        else:  # substitution
            act, node, cand = action.split(split_str)
            node = node
            cand = cand
            nodes = nodes + [cand]
            nodes = list(set(nodes).difference({node}))
        return nodes

    @staticmethod
    def to_key(nodes):
        """
        Creates a string representation of nodes
        :param nodes: node list
        :return: string of nodes
        """
        nodes = sorted(nodes)
        tmp_nodes = "|".join(nodes)
        return tmp_nodes

    def ls_on_genes(self, nodes, solutions, score0, T):
        """
        Runs local search on a gene set
        :param nodes: current node set
        :param solutions: dictionary wth previously used solutions
        :param score0: last objective function score
        :param T: temperature for SA

        :return:
        nodes - new set of nodes
        score1 - new score
        move - True if further optimization was possible
        """
        # SUBNETWORK OPTIMIZATION
        move = False  # indicates if the solution feasible
        AP = self.is_AP(nodes)
        results = {**self.insertion(nodes), **self.deletion(nodes, AP),
                   **self.subst(nodes, AP)}
        # first select the highest scoring solution which doesn't lead to the same set of nodes
        while not move:
            action = max(results.items(), key=operator.itemgetter(1))[0]
            score1 = results[action]
            # check if the solution is feasible
            nodes_new = self.do_action_nodes(action, nodes)
            nodes_new = self.to_key(nodes_new)
            if solutions.get(nodes_new) == None:  # solution wasn't used before
                move = True
            else:
                del results[action]
                if len(results) == 0:
                    if self.verbose:
                        print("no more feasible solutions")
                    return nodes, score0, move

        delta = score0 - score1
        if delta < 0:  # move on
            if self.verbose:
                print(action)
                print("Score after genes LS {0}".format(score1))
            nodes = self.do_action_nodes(action, nodes)

        else:  # SA
            try:
                val = np.exp(-delta / T)
            except RuntimeError:
                val = 0
            p = np.random.uniform()
            if val > p:  # move on
                if self.verbose:
                    print("SA on genes at {0} degrees".format(T))
                    print(action)
                    print("Score after genes LS".format(score1))
                nodes = self.do_action_nodes(action, nodes)
            else:  # terminate if no improvement in two rounds
                if self.verbose:

                    print("too cold for genes SA, no actions taken")
                move = False
                score1 = score0
        return nodes, score1, move

    def compute_seed(self):
        nodes = []
        # for  whatever reason dfs sometimes returns nothing
        no_type = True
        while no_type:
            nodes = self.dfs(np.random.choice(self.genes, 1)[0], self.init_size)
            if nodes is not None:
                no_type = False
        return nodes

    def run_ls(self):
        """
        Runs LS on patients and nodes
        :return:
        best_nodes - optimized node set
        score_max -maximal score
        """

        T0 = self.T
        T = T0
        score_max = -100
        best_nodes = []
        # initialization
        if self.seed is None:
            nodes = self.compute_seed()
        else:
            nodes = self.seed

        start_score = self.score(nodes)
        if self.verbose:
            print(start_score)
        score0 = start_score
        scores = [start_score]
        solutions = dict()
        nodes_keys = self.to_key(nodes)
        solutions[nodes_keys] = ""
        count = 0
        for it in range(self.max_iter):
            #print(it)
            score1 = score0
            nodes_backup = nodes
            nodes, score2, move_genes = self.ls_on_genes(nodes, solutions, score1, T)
            if not self.is_connected(nodes):
                print("something is wrong, network is disconnected")
                return nodes_backup, 0
            T = T0 * (0.9 ** it)  # cool down
            if self.verbose:
                print(it)
            solutions[self.to_key(nodes)] = ""
            scores.append(score2)

            score0 = score2
            if score2 > score_max:
                score_max = score2
                best_nodes = [x for x in nodes]

            count = count + 1
            if not move_genes:
                if self.plot:
                    convergence_plot(scores)
                break
        return best_nodes, score_max, scores

